"""Represent fhir entity."""

from anvil.transformers.fhir import make_identifier
import logging
from anvil.transformers.fhir.disease_normalizer import disease_text, disease_system

logged_already = []


class DiseaseObservation:
    """Create fhir entity."""

    class_name = "research_study"
    resource_type = "ResearchStudy"

    @staticmethod
    def build_entity(subject, disease):
        """Create FHIR entity."""
        assert disease, f'Should have disease {subject}'

        workspace_diseaseOntologyId = disease  # subject.workspace_diseaseOntologyId
        diseaseOntologyText = disease_text.get(disease, None)
        diseaseOntologySystem = disease_system.get(disease.split(':')[0], None)
        if workspace_diseaseOntologyId and not diseaseOntologyText:
            if workspace_diseaseOntologyId not in logged_already:
                logging.getLogger(__name__).error(f'Need text "{workspace_diseaseOntologyId}"')
                logged_already.append(workspace_diseaseOntologyId)
        assert diseaseOntologySystem, "Should have system"
        slug = make_identifier(f"Observation|{subject.id}|{workspace_diseaseOntologyId}")

        entity = {
            "resourceType": "Observation",
            "id": slug,
            "meta": {
                "profile": [
                    "http://fhir.ncpi-project-forge.io/StructureDefinition/ncpi-phenotype"
                ]
            },
            "identifier": [
                {
                    "system": "urn:ncpi:unique-string",
                    "value": slug
                }
            ],
            "status": "final",
            "code": {
                "coding": [
                    {
                        "system": diseaseOntologySystem,
                        "code": workspace_diseaseOntologyId,
                        "display": diseaseOntologyText
                    }
                ],
                "text": f"{diseaseOntologyText}"
            },
            "subject": {
                "reference": f"Patient/{make_identifier('P', subject.id)}"
            },
            "valueCodeableConcept": {
                "coding": [
                    {
                        "system": "http://snomed.info/sct",
                        "code": "373573001",
                        "display": "Clinical finding present (situation)"
                    }
                ],
                "text": "Phenotype Present"
            },
            "interpretation": [
                {
                    "coding": [
                        {
                            "system": "http://terminology.hl7.org/CodeSystem/v3-ObservationInterpretation",
                            "code": "POS",
                            "display": "Positive"
                        }
                    ],
                    "text": "Present"
                }
            ],
            "extension": []
        }
        if subject.age:
            entity['extension'].append(
                {
                    "url": "http://fhir.ncpi-project-forge.io/StructureDefinition/age-at-event",
                    "valueAge": {
                        "value": subject.age,
                        "unit": "y",
                        "system": "http://unitsofmeasure.org",
                        "code": "years"
                    }
                }
            )

        return entity



"""Client for reconciling terra subject."""

import logging
from attrdict import AttrDict


class Subject(object):
    """Represent terra subject."""

    def __init__(self, *args, workspace=None, samples=None):
        """Pass all args to AttrDict."""
        self.attributes = AttrDict(*args)
        self._logger = logging.getLogger(__name__)
        self.workspace_name = workspace.name
        self.schema = workspace.subject_schema
        self.sample_schema = workspace.sample_schema
        self.subject_property_name = workspace.subject_property_name
        self.namespace = workspace.attributes.workspace.namespace
        # find all samples associated with blobs
        self.samples = self._find_samples(samples)
        self.workspace_diseaseOntologyId = workspace.diseaseOntologyId

    def __repr__(self):
        """Return attributes."""
        return str(self.attributes)

    @property
    def id(self):
        """Delegate to subclass."""
        raise Exception('Not implemented')

    def _find_samples(self, samples):
        """Get samples."""
        return samples[self.attributes.name]

    @property
    def missing_samples(self):
        """Test if no samples."""
        return len(self.samples) == 0

    @property
    def age(self):
        """Delegate to subclass."""
        raise Exception('Not implemented')

    @property
    def gender(self):
        """Deduce gender."""
        for p in ['gender', 'sex']:
            if p in self.attributes.attributes:
                gender = self.attributes.attributes[p].lower()
                if gender in ['null', 'na', 'not reported', 'notreported']:
                    return None
                return gender.lower()
        logging.getLogger(__name__).info(f"{self.workspace_name} {self.id} missing gender parameter")
        return None

    @property
    def ethnicity(self):
        """Deduce ethnicity."""
        for p in ['11-ancestry_detail', '10-ancestry', 'ancestry', 'Race_Ethnicity', 'Ethnicity', 'RACE']:
            if p in self.attributes.attributes:
                ethnicity = self.attributes.attributes[p]
                if ethnicity in ['null', 'NA', '#N/A']:
                    return None
                if ethnicity in ['Hispanic or Latino', 'hispanic-or-latino', 'Hispanic', 'Hispanic/Latino', 'Puerto Rican']:
                    return 'hispanic'
                if ethnicity in ['African American', 'Black or African American']:
                    return 'black'
                if ethnicity in ['unknown', 'Unknown']:
                    return 'unknown'
                if ethnicity in ['White', 'Caucasian', 'Finnish']:
                    return 'white'
                if ethnicity in ['Not Hispanic or Latino', 'not-hispanic-or-latino', 'Non-Hispanic']:
                    return 'not-hispanic'
                if ethnicity in ['American Indian or Alaskan Native']:
                    return 'american-indian-or-alaskan-native'
                if ethnicity in ['not-asked', 'Not Asked']:
                    return 'not-asked'
                if ethnicity in ['Asian']:
                    return 'asian'
                return ethnicity
        logging.getLogger(__name__).info(f"{self.workspace_name} {self.id} missing ethnicity parameter")
        return None

    @property
    def phenotypes(self):
        """Deduce phenotype."""
        for k in ['hpo_present', '19-hpo_present', '21-phenotype_description']:
            if k in self.attributes.attributes:
                _s = self.attributes.attributes[k].replace(' ', '').replace(';', '|')
                return _s.split('|')
        return None

    @property
    def diseases(self):
        """Deduce diseases."""
        for k in ['disease_id', '14-disease_id']:
            if k in self.attributes.attributes:
                if not self.attributes.attributes[k] == '-':
                    _s = self.attributes.attributes[k].replace(' ', '').replace(';', '|')
                    return _s.split('|')
        # CCDG
        if 'Disease_Status' in self.attributes.attributes:
            if self.attributes.attributes.get('Disease_Status', '').lower() == 'case' and self.workspace_diseaseOntologyId:
                return [self.workspace_diseaseOntologyId]
        return []


class CCDGSubject(Subject):
    """Extend Subject class."""

    def __init__(self, *args, workspace=None, samples=None):
        """Call super."""
        super().__init__(*args, workspace=workspace, samples=samples)

    @property
    def id(self):
        """Deduce id."""
        return f"{self.workspace_name}/Su/{self.attributes.name}"

    @property
    def age(self):
        """Deduce age."""
        for p in ['Age', 'AGE', 'AGE_baseline']:
            if p in self.attributes.attributes:
                age = self.attributes.attributes[p]
                if not str(age).isnumeric():
                    logging.getLogger(__name__).warn(f"{self.workspace_name} {self.id} {p} not numeric '{age}'")
                    return None
                return int(age)
        logging.getLogger(__name__).info(f"{self.workspace_name} {self.id} missing age parameter")
        return None


class CMGSubject(Subject):
    """Extend Subject class."""

    def __init__(self, *args, workspace=None, samples=None):
        """Call super."""
        super().__init__(*args, workspace=workspace, samples=samples)

    @property
    def id(self):
        """Deduce id."""
        return f"{self.workspace_name}/Su/{self.attributes.name}"

    @property
    def age(self):
        """Deduce age."""
        for p in ['18-age_of_onset', "12-age_at_last_observation"]:
            if p in self.attributes.attributes:
                age = self.attributes.attributes[p]
                if not str(age).isnumeric():
                    if not age == '-':
                        logging.getLogger(__name__).warn(f"{self.workspace_name} {self.id} {p} not numeric '{age}'")
                    return None
                return int(age)
        logging.getLogger(__name__).info(f"{self.workspace_name} {self.id} missing age parameter")


class GTExSubject(Subject):
    """Extend Subject class."""

    def __init__(self, *args, workspace=None, samples=None):
        """Call super."""
        super().__init__(*args, workspace=workspace, samples=samples)

    @property
    def id(self):
        """Deduce id."""
        return f"{self.workspace_name}/Su/{self.attributes.name}"

    @property
    def age(self):
        """Deduce age."""
        if 'age' not in self.attributes.attributes:
            logging.getLogger(__name__).info(f"{self.workspace_name} {self.id} missing age")
            return None
        age = self.attributes.attributes['age']
        if not str(age).isnumeric():
            logging.getLogger(__name__).info(f"{self.workspace_name} {self.id} age not numeric")
            return None
        return int(age)


class ThousandGenomesSubject(Subject):
    """Extend Subject class."""

    def __init__(self, *args, workspace=None, samples=None):
        """Call super."""
        super().__init__(*args, workspace=workspace, samples=samples)

    @property
    def id(self):
        """Deduce id."""
        return f"{self.workspace_name}/Su/{self.attributes.name}"

    @property
    def age(self):
        """Deduce age."""
        return None


class eMERGESUbject(Subject):
    """Extend Subject class."""

    def __init__(self, *args, workspace=None, samples=None):
        """Call super."""
        super().__init__(*args, workspace=workspace, samples=samples)

    @property
    def id(self):
        """Deduce id."""
        return f"{self.workspace_name}/Su/{self.attributes.name}"

    @property
    def age(self):
        """Deduce age."""
        return None


def subject_factory(*args, **kwargs):
    """Return a specialized Subject class instance."""
    if 'CCDG' in kwargs['workspace'].name.upper():
        return CCDGSubject(*args, **kwargs)
    if 'CMG' in kwargs['workspace'].name.upper():
        return CMGSubject(*args, **kwargs)
    if 'GTEX' in kwargs['workspace'].name.upper():
        return GTExSubject(*args, **kwargs)
    if '1000G-HIGH-COVERAGE' in kwargs['workspace'].name.upper():
        return ThousandGenomesSubject(*args, **kwargs)
    if 'ANVIL_EMERGE' in kwargs['workspace'].name.upper():
        return eMERGESUbject(*args, **kwargs)
    raise Exception('Not implemented')



"""Validate AnVIL workspace(s)."""

import os

from google.cloud.storage import Client
from google.cloud.storage.blob import Blob
from collections import defaultdict
import ipywidgets as widgets
from ipywidgets import interact
from IPython.display import display
import pandas as pd
import firecloud.api as FAPI


class IngestHelper():
    """Validate workspace from dropdown selections."""

    def __init__(self, workspace_namespace='terra-test-bwalsh', workspace_name='pyAnVIL Notebook', user_project=os.environ.get('GOOGLE_PROJECT', None)) -> None:
        """Retrieve expected schemas."""
        assert user_project, f"AnVIL buckets use the `Requester Pays` feature. Please include a billing project."
        self.WORKSPACES = FAPI.list_workspaces().json()
        self.schemas_table = FAPI.get_entities(workspace_namespace, workspace_name, 'schema').json()
        self.schemas = defaultdict(dict)
        for e in self.schemas_table:
            a = e['attributes']
            self.schemas[a['consortium']][a['entity']] = a
        self.consortiums = widgets.Dropdown(options=['Choose...'] + list(self.schemas.keys()))
        self.workspaces = widgets.Dropdown(options=[])
        self.user_project = user_project
        self.client = Client(project=self.user_project)

    def validate(self, reference_schema, namespace, workspace_name, check_blobs=True):
        """Check target workspace against reference."""
        target_entities = FAPI.list_entity_types(namespace=namespace, workspace=workspace_name).json()
        reference = set(reference_schema.keys())
        target = set(target_entities.keys())
        result = dict(workspace=workspace_name)
        for entity in reference.intersection(target):
            uri = None
            try:
                reference_fields = set([f.replace(' ', '') for f in reference_schema[entity]['required'].split(',')])
                if 'bucket_fields' in reference_schema[entity]:
                    reference_fields.update([f.replace(' ', '') for f in reference_schema[entity].get('bucket_fields', '').split(',')])
                target_fields = set(target_entities[entity]['attributeNames'] + [target_entities[entity]['idName']])
                if not reference_fields.issubset(target_fields):
                    msg = f'fields_missing:{reference_fields - target_fields }'
                else:
                    msg = 'OK'
                result[entity] = msg
                project_buckets = {}
                if 'bucket_fields' in reference_schema[entity]:
                    for bucket_field in reference_schema[entity]['bucket_fields'].split(','):
                        if bucket_field not in target_fields:
                            result[entity] = f"{bucket_field} not found in {entity} schema."
                            continue
                        for e in FAPI.get_entities(namespace, workspace=workspace_name, etype=entity).json():
                            uri = e['attributes'][bucket_field]
                            blob = Blob.from_string(uri, client=self.client)
                            bucket_name = blob.bucket.name
                            if bucket_name not in project_buckets:
                                print(f"checking {workspace_name} {bucket_name}")
                                bucket = self.client.bucket(bucket_name, user_project=self.user_project)
                                project_buckets[bucket_name] = {}
                                for b in list(bucket.list_blobs()):
                                    project_buckets[bucket_name][b.name] = {'size': b.size, 'etag': b.etag, 'crc32c': b.crc32c, 'time_created': b.time_created, 'name': b.name}
                            if blob.name not in project_buckets[bucket_name]:
                                result[entity] = f"{uri} does not exist\n{e}\n{project_buckets[bucket_name].keys()}"
                                break
            except Exception as e:
                print(f"{workspace_name} {uri} {e}")
                result[entity] = str(e)

        for entity in reference - target:
            result[entity] = 'missing'
        result['unknown'] = f"{','.join(sorted(target - reference))}"

        return result

    def interact(self):
        """Use widgets to display drop downs for consortiums and workspaces, handle user selections."""
        pd.set_option("display.max_rows", None, "display.max_columns", None)

        def update_workspaces(*args):
            self.workspaces.options = ['Choose...', 'All workspaces'] + [w['workspace']['name'] for w in self.WORKSPACES if 'anvil-datastorage' in w['workspace']['namespace'] and self.consortiums.value.lower() in w['workspace']['name'].lower()]

        # Tie the image options to directory value
        self.consortiums.observe(update_workspaces, 'value')

        # Show the images
        def show_workspace(consortium, workspace):
            reference_df = pd.DataFrame(self.schemas[consortium]).dropna()
            reference_df = reference_df.style.set_caption("Reference")
            if workspace and workspace == 'All workspaces':
                print("Working...")
                validations = []
                for workspace_name in [w['workspace']['name'] for w in self.WORKSPACES if 'anvil-datastorage' in w['workspace']['namespace'] and self.consortiums.value.lower() in w['workspace']['name'].lower()]:
                    validation = self.validate(self.schemas[consortium], 'anvil-datastorage', workspace_name)
                    validations.append(validation)
                df = pd.DataFrame(validations).set_index('workspace').style.set_caption(f"{consortium}/{workspace}")
                display(reference_df)
                display(df)
                return

            if workspace and workspace != 'Choose...':
                df = pd.DataFrame([self.validate(self.schemas[consortium], 'anvil-datastorage', workspace)])
                df = df.set_index('workspace').style.set_caption(f"{consortium}/{workspace}")
                display(reference_df)
                display(df)
                return

        _ = interact(show_workspace, consortium=self.consortiums, workspace=self.workspaces)



"""Client for reconciling terra(google) blob."""

from attrdict import AttrDict


class Blob(object):
    """Represent terra(google) blob."""

    def __init__(self, blob, sample):
        """Simplify blob."""
        self.attributes = AttrDict(blob)
        self.sample = sample



"""Represent fhir entity."""


class Organization:
    """Create fhir entity."""

    class_name = "organization"
    resource_type = "Organization"

    @staticmethod
    def build_entity(workspace, parent=None):
        """Create fhir entity."""
        study_id = workspace.id
        id = study_id.lower()  # make_identifier(Organization.resource_type, study_id)
        # institution = workspace.institute
        # if not institution:
        #     logging.getLogger(__name__).warning(f'workspace {study_id} missing institute')
        #     institution = f"{study_id}-missing-institution"

        # workspace = workspace.attributes.workspace.attributes

        entity = {
            "resourceType": Organization.resource_type,
            "id": f"{id}",
            "meta": {
                "profile": [
                    "http://hl7.org/fhir/StructureDefinition/Organization"
                ]
            },
            "identifier": [
                {
                    "system": "https://anvil.terra.bio/#workspaces/anvil-datastorage/",
                    "value": study_id,
                },
                {
                    "system": "urn:ncpi:unique-string",
                    "value": f"{id}",
                },
            ],
            "name": study_id,
        }

        if parent:
            entity['partOf'] = {
                "reference": parent
            }
        elif 'attributes' in workspace:
            entity['partOf'] = {
                "reference": f"Organization/{workspace.attributes.reconciler_name.lower()}"
            }

        return entity



"""Utility methods."""
import re


def make_identifier(*args):
    """Create legal fhir id."""
    return re.sub(r"[^A-Za-z0-9\-\.]", "-", ".".join(str(a) for a in args))[-64:]


def join(*args):
    """Create pipe separated string."""
    return "|".join(str(a) for a in args)


# note coordinate with StructureDefintions /fhir/config.yaml::canonical
CANONICAL = "http://fhir.ncpi-project-forge.io"



"""Represent fhir entity."""

from anvil.transformers.fhir import CANONICAL, join, make_identifier
from anvil.transformers.fhir.practitioner import Practitioner
from anvil.transformers.fhir.organization import Organization

import logging

from anvil.transformers.fhir.disease_normalizer import disease_text, disease_system


class ResearchStudy:
    """Create fhir entity."""

    class_name = "research_study"
    resource_type = "ResearchStudy"

    @staticmethod
    def build_entity(workspace):
        """Create fhir entity."""
        study_id = workspace.id.lower()
        investigator_name = workspace.investigator
        if not investigator_name:
            logging.getLogger(__name__).warning(f'{study_id} missing investigator')
        diseaseOntologyId = workspace.diseaseOntologyId

        diseaseOntologyIdText = 'Missing'
        if diseaseOntologyId and diseaseOntologyId not in disease_text:
            logging.getLogger(__name__).error(f'{study_id} missing {diseaseOntologyId}')
        else:
            diseaseOntologyIdText = disease_text.get(diseaseOntologyId, 'Missing')

        workspace = workspace.attributes.workspace.attributes
        institution = workspace.get('institute', None)
        if not institution:
            logging.getLogger(__name__).warning(f'{study_id} missing institute')
        study_name = study_id
        attribution = study_id
        short_name = study_id
        key = study_id
        condition = None
        if diseaseOntologyId:
            prefix = diseaseOntologyId.split(':')[0]
            condition = [
                {
                    "coding": [
                        {
                            "system": disease_system[prefix],
                            "code": diseaseOntologyId,
                            "display": diseaseOntologyIdText,
                        }
                    ]
                }
            ]

        entity = {
            "resourceType": ResearchStudy.resource_type,
            "id": make_identifier(study_id),
            "meta": {
                "profile": [
                    "http://hl7.org/fhir/StructureDefinition/ResearchStudy"
                ]
            },
            "identifier": [
                {
                    "system": "https://anvil.terra.bio/#workspaces/anvil-datastorage/",
                    "value": study_id,
                },
                {
                    "system": "urn:ncpi:unique-string",
                    "value": join(ResearchStudy.resource_type, key),
                },
            ],
            "extension": [
                {
                    "url": f"{CANONICAL}/StructureDefinition/related-organization",
                    "extension": [
                        {
                            "url": "organization",
                            "valueReference": {
                                "reference": f"Organization/{make_identifier(Organization.resource_type, institution)}"
                            },
                        }
                    ],
                }
            ],
            "title": study_name,
            "status": "completed",
            "sponsor": {
                "reference": f"Organization/{study_id.lower()}"
            }
        }

        if investigator_name:
            entity["principalInvestigator"] = {
                "reference": f"Practitioner/{make_identifier(Practitioner.resource_type, investigator_name)}"
            }

        if attribution:
            entity["identifier"].append({"value": attribution})

        if short_name:
            entity["extension"].append(
                {
                    "url": f"{CANONICAL}/StructureDefinition/display-name",
                    "valueString": short_name,
                }
            )
        if condition:
            entity['condition'] = condition

        return entity



"""Parse PFB avro, write sqlite, summarize."""
from fastavro import reader
import json
from datetime import date, datetime
import sqlite3
import logging


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code."""
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError("Type %s not serializable" % type(obj))


class Entities:
    """Represent gen3 objects."""

    def __init__(self, path):
        """Simplify blob."""
        self.path = path
        self._conn = sqlite3.connect('/tmp/gen3-drs.sqlite')
        cur = self._conn.cursor()
        cur.executescript("""
        CREATE TABLE IF NOT EXISTS vertices (
            key text PRIMARY KEY,
            submitter_id text,
            name text,
            json text NOT NULL
        );
        CREATE TABLE IF NOT EXISTS edges (
            src text,
            dst text,
            src_name text,
            dst_name text
        );
        CREATE TABLE IF NOT EXISTS history (
            key text
        );
        """)
        self._conn.commit()

    def put(self, key, submitter_id, name, data, cur):
        """Save an item."""
        cur.execute("REPLACE into vertices values (?, ?, ?, ?);", (key, submitter_id, name, json.dumps(data, default=json_serial)))
        for r in data['relations']:
            cur.execute("REPLACE into edges values (?, ?, ?, ?);", (key, r['dst_id'], data['name'], r['dst_name']))
        # self._logger.debug(f"put {key}")

    def get(self, key=None, submitter_id=None):
        """Retrieve an item."""
        cur = self._conn.cursor()
        data = cur.execute("SELECT json FROM vertices where key=? or submitter_id =?", (key, submitter_id)).fetchone()
        if data:
            return json.loads(data[0])
        assert False, f"NOT FOUND {key} {submitter_id}"

    def load(self):
        """Load sqlite db from file."""
        cur = self._conn.cursor()
        ids = {}
        logging.getLogger(__name__).info(f'Loading {self.path}')

        loaded_already = cur.execute("SELECT count(*) FROM history WHERE key=?;", (self.path,)).fetchone()[0]
        if loaded_already == 1:
            logging.getLogger(__name__).info(f'Already indexed {self.path}')
            return

        # index_count = cur.execute("SELECT count(*) FROM sqlite_master WHERE type='index' and name='vertices_submitter_id';").fetchone()[0]
        # if index_count == 1:
        #     logging.getLogger(__name__).info('Already indexed')
        #     return

        with open(self.path, 'rb') as fo:
            for record in reader(fo):
                if record['id'] not in ids:
                    self.put(record['id'], self._submitter_id(record), record['name'], record, cur)
                    ids[record['id']] = None
        self._conn.commit()

        logging.getLogger(__name__).info('Indexing')
        cur.executescript("""
        CREATE  INDEX IF NOT EXISTS vertices_submitter_id ON vertices(submitter_id);
        CREATE UNIQUE INDEX IF NOT EXISTS edges_src_dst ON edges(src, dst, src_name, dst_name);
        CREATE  INDEX IF NOT EXISTS edges_dst ON edges(dst);
        """)
        self._conn.commit()

        logging.getLogger(__name__).info('Flattening')
        cur.executescript("""
        drop table if exists flattened ;
        create table flattened
        as
        select
            json_extract(su.json, '$.object.project_id') as "project_id",
            json_extract(su.json, '$.object.anvil_project_id') as "anvil_project_id",
            su.name as "subject_type",
            su.key as "subject_id",
            json_extract(su.json, '$.object.participant_id') as "participant_id",
            json_extract(su.json, '$.object.submitter_id') as "subject_submitter_id",
            sa.name as "sample_type",
            sa.key  as "sample_id",
            json_extract(sa.json, '$.object.sample_id') as "sample_sample_id",
            json_extract(sa.json, '$.object.submitter_id') as "sample_submitter_id",
            'sequencing' as "sequencing_type",
            sequencing_edge.src  as "sequencing_id",
            json_extract(sq.json, '$.object.submitter_id') as "sequencing_submitter_id",
            json_extract(sq.json, '$.object.ga4gh_drs_uri') as "ga4gh_drs_uri"
            from vertices as su
                join edges as sample_edge on sample_edge.dst = su.key and sample_edge.src_name = 'sample'
                    join vertices as sa on sample_edge.src = sa.key
                        left join edges as sequencing_edge on sequencing_edge.dst = sa.key and sequencing_edge.src_name = 'sequencing'
                            join vertices as sq on sequencing_edge.src = sq.key

            where
            su.name = 'subject'            ;
        """)
        self._conn.commit()

        logging.getLogger(__name__).info('Summarizing')
        cur.executescript("""
            drop table if exists summary ;
            create table summary
            as
            select f.project_id, f.anvil_project_id,
                count(distinct f.subject_id) as "subject_count",
                count(distinct f.sample_id) as "sample_count",
                count(distinct m.sequencing_id) as "sequencing_count",
                count(distinct m.ga4gh_drs_uri) as "ga4gh_drs_uri_count"
                from flattened as f
                    left join flattened as m on f.project_id = m.project_id and f.anvil_project_id = m.anvil_project_id
                group by f.project_id, f.anvil_project_id;
        """)
        self._conn.commit()

        logging.getLogger(__name__).info('Updating history')
        cur.execute("""
            insert into history(key) values(?);
        """, (self.path,))
        self._conn.commit()

    def _submitter_id(self, record):
        """Deduce 'natural' key.

        See https://docs.google.com/spreadsheets/d/1MxfcWDXhTfFNFKsbRGjGTQkBoTirNktj04lf6L9_jmk/edit#gid=0
        """
        name = record['name']
        if 'specimen_id' in record['object'] and name == 'sample':
            return record['object']['specimen_id']
        if 'participant_id' in record['object'] and name == 'subject':
            return record['object']['participant_id']
        if 'file_name' in record['object'] and name == 'sequencing':
            return record['object']['file_name']
        if 'dbgap_accession_number' in record['object'] and name == 'program':
            return record['object']['dbgap_accession_number']
        if 'code' in record['object'] and name == 'project':
            return record['object']['code']
        if record['name'] == 'Metadata':
            return f"{record['name']}/schema"
        if 'submitter_id' in record['object']:
            return record['object']['submitter_id']
        assert False, record



"""Wraps  firecloud api."""

from anvil.util.cache import memoize
import firecloud.api as FAPI
import logging
import re
from attrdict import AttrDict

logger = logging.getLogger('anvil.terra.api')

USER_PROJECT = None


def whoami():
    """Wrap fapi's whoami.

    Returns:
        str: google id

    """
    return FAPI.whoami()


@memoize
def get_projects(namespaces=None, project_pattern=None):
    """Filter terra workspaces by namespaces and project_pattern.

    Args:
        namespaces ([str]): Optional, list of workspace `namespace` to match ex: 'anvil-datastorage'.
        project_pattern (str): Optional, regexp to match workspace `name` ex: 'AnVIL_CCDG.*'.

    Returns:
        dict: keys ['accessLevel', 'public', 'workspace', 'workspaceSubmissionStats']

    """
    workspaces = FAPI.list_workspaces()
    workspaces = workspaces.json()

    if namespaces:
        workspaces = [AttrDict(w) for w in workspaces if w['workspace']['namespace'] in namespaces]

    if project_pattern:
        workspaces = [AttrDict(w) for w in workspaces if re.match(project_pattern, w['workspace']['name'], re.IGNORECASE)]

    # normalize fields
    for w in workspaces:
        if 'project_files' not in w.workspace:
            w.workspace.project_files = []
    return workspaces


@memoize
def get_entities(namespace='anvil-datastorage', workspace=None, entity_name=None):
    """Return all entities in a workspace."""
    entities = [AttrDict(e) for e in FAPI.get_entities(namespace, workspace, entity_name).json()]
    return entities


@memoize
def get_schema(namespace, workspace):
    """Fetch all entity types."""
    return FAPI.list_entity_types(namespace=namespace, workspace=workspace).json()



"""Dictionaries supporting conditions and phenotypes."""

disease_text = {
    "DOID:0050589": "inflammatory bowel disease",
    "DOID:9744": "type 1 diabetes mellitus",
    "DOID:1287": "cardiovascular system disease",
    "DOID:3393": "coronary artery disease",
    "DOID:0060224": "atrial fibrillation",
    "DOID:0014667": "disease of metabolism",
    "DOID:10652": "Alzheimer's disease",
    "DOID:0060041": "autism spectrum disorder",
    "DOID:2841": "asthma",
    "DOID:5844": "myocardial infarction",
    "DOID:1826": "epilepsy",
    "DOID:630": "genetic disease",
    "OMIM:613801": "Retinitis pigmentosa-40",
    "OMIM:213300": "Joubert syndrome 1",
    "OMIM:615780": "Retinitis pigmentosa 69",
    "OMIM:613826": "Leber congenital amaurosis 6",
    "OMIM:310200": "Duchenne muscular dystrophy",
    "OMIM:617194": "Lethal congenital contracture syndrome 11",
    "OMIM:105400": "AMYOTROPHIC LATERAL SCLEROSIS 1; ALS1",
    "OMIM:108120": "ARTHROGRYPOSIS, DISTAL, TYPE 1A; DA1A",
    "OMIM:114300": "ARTHROGRYPOSIS, DISTAL, TYPE 3; DA3",
    "OMIM:115195": "CARDIOMYOPATHY, FAMILIAL HYPERTROPHIC, 2; CMH2",
    "OMIM:115196": "CARDIOMYOPATHY, FAMILIAL HYPERTROPHIC, 3; CMH3",
    "OMIM:115197": "CARDIOMYOPATHY, FAMILIAL HYPERTROPHIC, 4; CMH4",
    "OMIM:115200": "CARDIOMYOPATHY, DILATED, 1A; CMD1A",
    "OMIM:117000": "CENTRAL CORE DISEASE OF MUSCLE; CCD",
    "OMIM:117360": "SPINOCEREBELLAR ATAXIA 29; SCA29",
    "OMIM:119530": "OROFACIAL CLEFT 1; OFC1",
    "OMIM:122470": "CORNELIA DE LANGE SYNDROME 1; CDLS1",
    "OMIM:130600": "ELLIPTOCYTOSIS 2; EL2",
    "OMIM:135900": "COFFIN-SIRIS SYNDROME 1; CSS1",
    "OMIM:136140": "FLOATING-HARBOR SYNDROME; FLHS",
    "OMIM:137920": "RENAL CYSTS AND DIABETES SYNDROME; RCAD",
    "OMIM:139090": "GRAY PLATELET SYNDROME; GPS",
    "OMIM:139210": "MYHRE SYNDROME; MYHRS",
    "OMIM:143095": "SPONDYLOEPIPHYSEAL DYSPLASIA WITH CONGENITAL JOINT DISLOCATIONS; SEDCJD",
    "OMIM:143465": "ATTENTION DEFICIT-HYPERACTIVITY DISORDER; ADHD",
    "OMIM:147891": "ISCHIOCOXOPODOPATELLAR SYNDROME WITH OR WITHOUT PULMONARY ARTERIAL HYPERTENSION; ICPPS",
    "OMIM:147920": "KABUKI SYNDROME 1; KABUK1",
    "OMIM:148050": "KBG SYNDROME; KBGS",
    "OMIM:149400": "HYPEREKPLEXIA 1; HKPX1",
    "OMIM:152950": "MICROCEPHALY WITH OR WITHOUT CHORIORETINOPATHY, LYMPHEDEMA, OR MENTAL RETARDATION; MCLMR",
    "OMIM:153670": "BERNARD-SOULIER SYNDROME, TYPE A2, AUTOSOMAL DOMINANT; BSSA2",
    "OMIM:154400": "ACROFACIAL DYSOSTOSIS 1, NAGER TYPE; AFD1",
    "OMIM:154700": "MARFAN SYNDROME; MFS",
    "OMIM:155310": "VISCERAL MYOPATHY; VSCM",
    "OMIM:158590": "NEURONOPATHY, DISTAL HEREDITARY MOTOR, TYPE IIA; HMN2A",
    "OMIM:158810": "BETHLEM MYOPATHY 1; BTHLM1",
    "OMIM:159550": "ATAXIA-PANCYTOPENIA SYNDROME; ATXPC",
    "OMIM:160150": "MYOPATHY, CENTRONUCLEAR, 1; CNM1",
    "OMIM:160500": "MYOPATHY, DISTAL, 1; MPD1",
    "OMIM:160800": "MYOTONIA CONGENITA, AUTOSOMAL DOMINANT",
    "OMIM:160900": "MYOTONIC DYSTROPHY 1; DM1",
    "OMIM:161800": "NEMALINE MYOPATHY 3; NEM3",
    "OMIM:163950": "NOONAN SYNDROME 1; NS1",
    "OMIM:178110": "CONTRACTURES, PTERYGIA, AND SPONDYLOCARPOTARSAL FUSION SYNDROME 1A; CPSFS1A",
    "OMIM:178500": "PULMONARY FIBROSIS, IDIOPATHIC; IPF",
    "OMIM:180100": "RETINITIS PIGMENTOSA 1; RP1",
    "OMIM:180849": "RUBINSTEIN-TAYBI SYNDROME 1; RSTS1",
    "OMIM:181430": "SCAPULOPERONEAL MYOPATHY, MYH7-RELATED; SPMM",
    "OMIM:182601": "SPASTIC PARAPLEGIA 4, AUTOSOMAL DOMINANT; SPG4",
    "OMIM:188000": "THROMBOCYTOPENIA 2; THC2",
    "OMIM:191480": "UNCOMBABLE HAIR SYNDROME 1; UHS1",
    "OMIM:203800": "ALSTROM SYNDROME; ALMS",
    "OMIM:204200": "CEROID LIPOFUSCINOSIS, NEURONAL, 3; CLN3",
    "OMIM:204500": "CEROID LIPOFUSCINOSIS, NEURONAL, 2; CLN2",
    "OMIM:205100": "AMYOTROPHIC LATERAL SCLEROSIS 2, JUVENILE; ALS2",
    "OMIM:206900": "MICROPHTHALMIA, SYNDROMIC 3; MCOPS3",
    "OMIM:208900": "ATAXIA-TELANGIECTASIA; AT",
    "OMIM:209850": "AUTISM",
    "OMIM:209900": "BARDET-BIEDL SYNDROME 1; BBS1",
    "OMIM:210000": "BEHR SYNDROME; BEHRS",
    "OMIM:210720": "MICROCEPHALIC OSTEODYSPLASTIC PRIMORDIAL DWARFISM, TYPE II; MOPD2",
    "OMIM:210900": "BLOOM SYNDROME; BLM",
    "OMIM:212065": "CONGENITAL DISORDER OF GLYCOSYLATION, TYPE Ia; CDG1A",
    "OMIM:212066": "CONGENITAL DISORDER OF GLYCOSYLATION, TYPE IIa; CDG2A",
    "OMIM:214500": "CHEDIAK-HIGASHI SYNDROME; CHS",
    "OMIM:214800": "CHARGE SYNDROME",
    "OMIM:215500": "CHOROIDAL DYSTROPHY, CENTRAL AREOLAR, 1; CACD1",
    "OMIM:216400": "COCKAYNE SYNDROME A; CSA",
    "OMIM:216550": "COHEN SYNDROME; COH1",
    "OMIM:217080": "JALILI SYNDROME",
    "OMIM:217980": "CORPUS CALLOSUM, AGENESIS OF, WITH FACIAL ANOMALIES AND ROBIN SEQUENCE",
    "OMIM:218340": "TEMTAMY SYNDROME; TEMTYS",
    "OMIM:219200": "CUTIS LAXA, AUTOSOMAL RECESSIVE, TYPE IIA; ARCL2A",
    "OMIM:224050": "CEREBELLAR ATAXIA, MENTAL RETARDATION, AND DYSEQUILIBRIUM SYNDROME 1; CAMRQ1",
    "OMIM:224690": "MEIER-GORLIN SYNDROME 1; MGORS1",
    "OMIM:225753": "PONTOCEREBELLAR HYPOPLASIA, TYPE 4; PCH4",
    "OMIM:225790": "PROLIFERATIVE VASCULOPATHY AND HYDRANENCEPHALY-HYDROCEPHALY SYNDROME; PVHH",
    "OMIM:227650": "FANCONI ANEMIA, COMPLEMENTATION GROUP A; FANCA",
    "OMIM:229600": "FRUCTOSE INTOLERANCE, HEREDITARY; HFI",
    "OMIM:230000": "FUCOSIDOSIS",
    "OMIM:230500": "GM1-GANGLIOSIDOSIS, TYPE I",
    "OMIM:232300": "GLYCOGEN STORAGE DISEASE II; GSD2",
    "OMIM:232600": "GLYCOGEN STORAGE DISEASE V; GSD5",
    "OMIM:234200": "NEURODEGENERATION WITH BRAIN IRON ACCUMULATION 1; NBIA1",
    "OMIM:234500": "HARTNUP DISORDER; HND",
    "OMIM:236000": "LYMPHOMA, HODGKIN, CLASSIC; CHL",
    "OMIM:236600": "HYDROCEPHALUS, CONGENITAL, 1; HYC1",
    "OMIM:236792": "L-2-HYDROXYGLUTARIC ACIDURIA; L2HGA",
    "OMIM:243310": "BARAITSER-WINTER SYNDROME 1; BRWS1",
    "OMIM:244450": "KAUFMAN OCULOCEREBROFACIAL SYNDROME; KOS",
    "OMIM:245200": "KRABBE DISEASE",
    "OMIM:245359": "missing",
    "OMIM:248200": "STARGARDT DISEASE 1; STGD1",
    "OMIM:248700": "MARDEN-WALKER SYNDROME; MWKS",
    "OMIM:249000": "MECKEL SYNDROME, TYPE 1; MKS1",
    "OMIM:249900": "METACHROMATIC LEUKODYSTROPHY DUE TO SAPOSIN B DEFICIENCY",
    "OMIM:250100": "METACHROMATIC LEUKODYSTROPHY; MLD",
    "OMIM:250620": "3-HYDROXYISOBUTYRYL-CoA HYDROLASE DEFICIENCY; HIBCHD",
    "OMIM:251200": "MICROCEPHALY 1, PRIMARY, AUTOSOMAL RECESSIVE; MCPH1",
    "OMIM:251260": "NIJMEGEN BREAKAGE SYNDROME; NBS",
    "OMIM:251280": "DIENCEPHALIC-MESENCEPHALIC JUNCTION DYSPLASIA SYNDROME 1; DMJDS1",
    "OMIM:251290": "PSEUDO-TORCH SYNDROME 1; PTORCH1",
    "OMIM:251300": "GALLOWAY-MOWAT SYNDROME 1; GAMOS1",
    "OMIM:252010": "MITOCHONDRIAL COMPLEX I DEFICIENCY, NUCLEAR TYPE 1; MC1DN1",
    "OMIM:252160": "MOLYBDENUM COFACTOR DEFICIENCY, COMPLEMENTATION GROUP B; MOCODB",
    "OMIM:252920": "MUCOPOLYSACCHARIDOSIS, TYPE IIIB; MPS3B",
    "OMIM:253250": "MULIBREY NANISM",
    "OMIM:253600": "MUSCULAR DYSTROPHY, LIMB-GIRDLE, AUTOSOMAL RECESSIVE 1; LGMDR1",
    "OMIM:253601": "MUSCULAR DYSTROPHY, LIMB-GIRDLE, AUTOSOMAL RECESSIVE 2; LGMDR2",
    "OMIM:253700": "MUSCULAR DYSTROPHY, LIMB-GIRDLE, AUTOSOMAL RECESSIVE 5; LGMDR5",
    "OMIM:254090": "ULLRICH CONGENITAL MUSCULAR DYSTROPHY 1; UCMD1",
    "OMIM:254110": "MUSCULAR DYSTROPHY, LIMB-GIRDLE, AUTOSOMAL RECESSIVE 8; LGMDR8",
    "OMIM:254300": "MYASTHENIC SYNDROME, CONGENITAL, 10; CMS10",
    "OMIM:254780": "MYOCLONIC EPILEPSY OF LAFORA",
    "OMIM:254940": "CAREY-FINEMAN-ZITER SYNDROME; CFZS",
    "OMIM:255310": "MYOPATHY, CONGENITAL, WITH FIBER-TYPE DISPROPORTION; CFTD",
    "OMIM:255320": "MINICORE MYOPATHY WITH EXTERNAL OPHTHALMOPLEGIA",
    "OMIM:255800": "SCHWARTZ-JAMPEL SYNDROME, TYPE 1; SJS1",
    "OMIM:256000": "LEIGH SYNDROME; LS",
    "OMIM:256300": "NEPHROTIC SYNDROME, TYPE 1; NPHS1",
    "OMIM:256600": "NEURODEGENERATION WITH BRAIN IRON ACCUMULATION 2A; NBIA2A",
    "OMIM:256700": "NEUROBLASTOMA, SUSCEPTIBILITY TO, 1; NBLST1",
    "OMIM:256730": "CEROID LIPOFUSCINOSIS, NEURONAL, 1; CLN1",
    "OMIM:256731": "CEROID LIPOFUSCINOSIS, NEURONAL, 5; CLN5",
    "OMIM:256800": "INSENSITIVITY TO PAIN, CONGENITAL, WITH ANHIDROSIS; CIPA",
    "OMIM:256850": "GIANT AXONAL NEUROPATHY 1, AUTOSOMAL RECESSIVE; GAN1",
    "OMIM:257220": "NIEMANN-PICK DISEASE, TYPE C1; NPC1",
    "OMIM:257320": "LISSENCEPHALY 2; LIS2",
    "OMIM:260400": "SHWACHMAN-DIAMOND SYNDROME 1; SDS1",
    "OMIM:260565": "PEHO SYNDROME; PEHO",
    "OMIM:261630": "HYPERPHENYLALANINEMIA, BH4-DEFICIENT, C; HPABH4C",
    "OMIM:262500": "LARON SYNDROME",
    "OMIM:263200": "POLYCYSTIC KIDNEY DISEASE 4 WITH OR WITHOUT POLYCYSTIC LIVER DISEASE; PKD4",
    "OMIM:265000": "MULTIPLE PTERYGIUM SYNDROME, ESCOBAR VARIANT; EVMPS",
    "OMIM:266100": "EPILEPSY, PYRIDOXINE-DEPENDENT; EPD",
    "OMIM:266265": "CONGENITAL DISORDER OF GLYCOSYLATION, TYPE IIc; CDG2C",
    "OMIM:266500": "REFSUM DISEASE, CLASSIC",
    "OMIM:266510": "PEROXISOME BIOGENESIS DISORDER 3B; PBD3B",
    "OMIM:268800": "SANDHOFF DISEASE",
    "OMIM:270200": "SJOGREN-LARSSON SYNDROME; SLS",
    "OMIM:270800": "SPASTIC PARAPLEGIA 5A, AUTOSOMAL RECESSIVE; SPG5A",
    "OMIM:271900": "CANAVAN DISEASE",
    "OMIM:272800": "TAY-SACHS DISEASE; TSD",
    "OMIM:273300": "TESTICULAR GERM CELL TUMOR; TGCT",
    "OMIM:274000": "THROMBOCYTOPENIA-ABSENT RADIUS SYNDROME; TAR",
    "OMIM:276901": "USHER SYNDROME, TYPE IIA; USH2A",
    "OMIM:277470": "PONTOCEREBELLAR HYPOPLASIA, TYPE 2A; PCH2A",
    "OMIM:300029": "RETINITIS PIGMENTOSA 3; RP3",
    "OMIM:300067": "LISSENCEPHALY, X-LINKED, 1; LISX1",
    "OMIM:300088": "DEVELOPMENTAL AND EPILEPTIC ENCEPHALOPATHY 9; DEE9",
    "OMIM:300100": "ADRENOLEUKODYSTROPHY; ALD",
    "OMIM:300376": "MUSCULAR DYSTROPHY, BECKER TYPE; BMD",
    "OMIM:300387": "MENTAL RETARDATION, X-LINKED 63; MRX63",
    "OMIM:300476": "CONE-ROD DYSTROPHY, X-LINKED, 3; CORDX3",
    "OMIM:300486": "MENTAL RETARDATION, X-LINKED, WITH CEREBELLAR HYPOPLASIA AND DISTINCTIVE FACIAL APPEARANCE",
    "OMIM:300495": "AUTISM, SUSCEPTIBILITY TO, X-LINKED 2; AUTSX2",
    "OMIM:300523": "ALLAN-HERNDON-DUDLEY SYNDROME; AHDS",
    "OMIM:300534": "MENTAL RETARDATION, X-LINKED, SYNDROMIC, CLAES-JENSEN TYPE; MRXSCJ",
    "OMIM:300695": "SCAPULOPERONEAL MYOPATHY, X-LINKED DOMINANT; SPM",
    "OMIM:300749": "MENTAL RETARDATION AND MICROCEPHALY WITH PONTINE AND CEREBELLAR HYPOPLASIA; MICPCH",
    "OMIM:300804": "JOUBERT SYNDROME 10; JBTS10",
    "OMIM:300881": "MOVED TO 615777",
    "OMIM:300882": "CORNELIA DE LANGE SYNDROME 5; CDLS5",
    "OMIM:300894": "NEURODEGENERATION WITH BRAIN IRON ACCUMULATION 5; NBIA5",
    "OMIM:300908": "ANEMIA, NONSPHEROCYTIC HEMOLYTIC, DUE TO G6PD DEFICIENCY",
    "OMIM:300957": "MENTAL RETARDATION, X-LINKED 12; MRX12",
    "OMIM:300958": "INTELLECTUAL DEVELOPMENTAL DISORDER, X-LINKED, SYNDROMIC, SNIJDERS BLOK TYPE; MRXSSB",
    "OMIM:300966": "MENTAL RETARDATION, X-LINKED, SYNDROMIC 33; MRXS33",
    "OMIM:300968": "MENTAL RETARDATION, X-LINKED 99, SYNDROMIC, FEMALE-RESTRICTED; MRXS99F",
    "OMIM:300983": "MENTAL RETARDATION, X-LINKED 104; MRX104",
    "OMIM:300986": "MENTAL RETARDATION, X-LINKED, SYNDROMIC, BAIN TYPE; MRXSB",
    "OMIM:301050": "ALPORT SYNDROME 1, X-LINKED; ATS1",
    "OMIM:301310": "ANEMIA, SIDEROBLASTIC, AND SPINOCEREBELLAR ATAXIA; ASAT",
    "OMIM:304340": "PETTIGREW SYNDROME; PGS",
    "OMIM:305100": "ECTODERMAL DYSPLASIA 1, HYPOHIDROTIC, X-LINKED; XHED",
    "OMIM:308350": "DEVELOPMENTAL AND EPILEPTIC ENCEPHALOPATHY 1; DEE1",
    "OMIM:309500": "RENPENNING SYNDROME 1; RENS1",
    "OMIM:309541": "METHYLMALONIC ACIDEMIA AND HOMOCYSTEINEMIA, cblX TYPE",
    "OMIM:309580": "MENTAL RETARDATION-HYPOTONIC FACIES SYNDROME, X-LINKED, 1; MRXHF1",
    "OMIM:310468": "NEPHROLITHIASIS, X-LINKED RECESSIVE, WITH RENAL FAILURE; XRN",
    "OMIM:311070": "CHARCOT-MARIE-TOOTH DISEASE, X-LINKED RECESSIVE, 5; CMTX5",
    "OMIM:313850": "THORACOABDOMINAL SYNDROME; THAS",
    "OMIM:314580": "WIEACKER-WOLFF SYNDROME; WRWF",
    "OMIM:600059": "RETINITIS PIGMENTOSA 13; RP13",
    "OMIM:600105": "RETINITIS PIGMENTOSA 12; RP12",
    "OMIM:600118": "WARBURG MICRO SYNDROME 1; WARBM1",
    "OMIM:600132": "RETINITIS PIGMENTOSA 14; RP14",
    "OMIM:600138": "RETINITIS PIGMENTOSA 11; RP11",
    "OMIM:600143": "CEROID LIPOFUSCINOSIS, NEURONAL, 8; CLN8",
    "OMIM:600334": "TIBIAL MUSCULAR DYSTROPHY, TARDIVE; TMD",
    "OMIM:600373": "CODAS SYNDROME",
    "OMIM:600618": "ETS VARIANT GENE 6; ETV6",
    "OMIM:600995": "NEPHROTIC SYNDROME, TYPE 2; NPHS2",
    "OMIM:601088": "AYME-GRIPP SYNDROME; AYGRP",
    "OMIM:601414": "RETINITIS PIGMENTOSA 18; RP18",
    "OMIM:601419": "MYOPATHY, MYOFIBRILLAR, 1; MFM1",
    "OMIM:601680": "ARTHROGRYPOSIS, DISTAL, TYPE 2B1; DA2B1",
    "OMIM:601718": "RETINITIS PIGMENTOSA 19; RP19",
    "OMIM:601780": "CEROID LIPOFUSCINOSIS, NEURONAL, 6; CLN6",
    "OMIM:601815": "PHOSPHOGLYCERATE DEHYDROGENASE DEFICIENCY; PHGDHD",
    "OMIM:601954": "MUSCULAR DYSTROPHY, LIMB-GIRDLE, AUTOSOMAL RECESSIVE 7; LGMDR7",
    "OMIM:602482": "AXENFELD-RIEGER SYNDROME, TYPE 3; RIEG3",
    "OMIM:602579": "CONGENITAL DISORDER OF GLYCOSYLATION, TYPE Ib; CDG1B",
    "OMIM:602772": "RETINITIS PIGMENTOSA 25; RP25",
    "OMIM:603511": "MUSCULAR DYSTROPHY, LIMB-GIRDLE, AUTOSOMAL DOMINANT 1; LGMDD1",
    "OMIM:603896": "LEUKOENCEPHALOPATHY WITH VANISHING WHITE MATTER; VWM",
    "OMIM:604004": "MEGALENCEPHALIC LEUKOENCEPHALOPATHY WITH SUBCORTICAL CYSTS 1; MLC1",
    "OMIM:604116": "CONE-ROD DYSTROPHY 3; CORD3",
    "OMIM:604273": "MITOCHONDRIAL COMPLEX V (ATP SYNTHASE) DEFICIENCY, NUCLEAR TYPE 1; MC5DN1",
    "OMIM:604286": "MUSCULAR DYSTROPHY, LIMB-GIRDLE, AUTOSOMAL RECESSIVE 4; LGMDR4",
    "OMIM:604317": "MICROCEPHALY 2, PRIMARY, AUTOSOMAL RECESSIVE, WITH OR WITHOUT CORTICAL MALFORMATIONS; MCPH2",
    "OMIM:604360": "SPASTIC PARAPLEGIA 11, AUTOSOMAL RECESSIVE; SPG11",
    "OMIM:604387": "NEPHRONOPHTHISIS 3; NPHP3",
    "OMIM:604537": "LEBER CONGENITAL AMAUROSIS 5; LCA5",
    "OMIM:604592": "T CELL IMMUNE REGULATOR 1; TCIRG1",
    "OMIM:605130": "WIEDEMANN-STEINER SYNDROME; WDSTS",
    "OMIM:605355": "NEMALINE MYOPATHY 5; NEM5",
    "OMIM:605407": "SEGAWA SYNDROME, AUTOSOMAL RECESSIVE",
    "OMIM:605637": "MYOPATHY, PROXIMAL, WITH OPHTHALMOPLEGIA; MYPOP",
    "OMIM:605670": "LATE-ONSET RETINAL DEGENERATION; LORD",
    "OMIM:605820": "NONAKA MYOPATHY; NM",
    "OMIM:605899": "GLYCINE ENCEPHALOPATHY; GCE",
    "OMIM:606068": "RETINITIS PIGMENTOSA 28; RP28",
    "OMIM:606170": "GENITOPATELLAR SYNDROME; GTPTS",
    "OMIM:606232": "PHELAN-MCDERMID SYNDROME; PHMDS",
    "OMIM:606777": "GLUT1 DEFICIENCY SYNDROME 1; GLUT1DS1",
    "OMIM:606854": "POLYMICROGYRIA, BILATERAL FRONTOPARIETAL; BFPP",
    "OMIM:606966": "NEPHRONOPHTHISIS 4; NPHP4",
    "OMIM:606996": "SENIOR-LOKEN SYNDROME 4; SLSN4",
    "OMIM:607015": "HURLER-SCHEIE SYNDROME",
    "OMIM:607155": "MUSCULAR DYSTROPHY-DYSTROGLYCANOPATHY (LIMB-GIRDLE), TYPE C, 5; MDDGC5",
    "OMIM:607196": "MICROCEPHALY, AMISH TYPE; MCPHA",
    "OMIM:607225": "SPASTIC PARALYSIS, INFANTILE-ONSET ASCENDING; IAHSP",
    "OMIM:607426": "COENZYME Q10 DEFICIENCY, PRIMARY, 1; COQ10D1",
    "OMIM:607432": "LISSENCEPHALY 1; LIS1",
    "OMIM:607483": "THIAMINE METABOLISM DYSFUNCTION SYNDROME 2 (BIOTIN- OR THIAMINE-RESPONSIVE TYPE); THMD2",
    "OMIM:607595": "MOVED TO 175780",
    "OMIM:607596": "PONTOCEREBELLAR HYPOPLASIA, TYPE 1A; PCH1A",
    "OMIM:607855": "MUSCULAR DYSTROPHY, CONGENITAL MEROSIN-DEFICIENT, 1A; MDC1A",
    "OMIM:608027": "PONTOCEREBELLAR HYPOPLASIA, TYPE 3; PCH3",
    "OMIM:608091": "JOUBERT SYNDROME 2; JBTS2",
    "OMIM:608099": "MUSCULAR DYSTROPHY, LIMB-GIRDLE, AUTOSOMAL RECESSIVE 3; LGMDR3",
    "OMIM:608133": "RETINITIS PIGMENTOSA 7; RP7",
    "OMIM:608358": "MYOPATHY, MYOSIN STORAGE, AUTOSOMAL DOMINANT; MSMA",
    "OMIM:608380": "RETINITIS PIGMENTOSA 26; RP26",
    "OMIM:608393": "MICROCEPHALY 6, PRIMARY, AUTOSOMAL RECESSIVE; MCPH6",
    "OMIM:608415": "PROLONGED ELECTRORETINAL RESPONSE SUPPRESSION; PERRS",
    "OMIM:608423": "MUSCULAR DYSTROPHY, LIMB-GIRDLE, AUTOSOMAL DOMINANT 2; LGMDD2",
    "OMIM:608540": "CONGENITAL DISORDER OF GLYCOSYLATION, TYPE Ik; CDG1K",
    "OMIM:608629": "JOUBERT SYNDROME 3; JBTS3",
    "OMIM:608643": "AROMATIC L-AMINO ACID DECARBOXYLASE DEFICIENCY",
    "OMIM:608716": "MICROCEPHALY 5, PRIMARY, AUTOSOMAL RECESSIVE; MCPH5",
    "OMIM:608804": "LEUKODYSTROPHY, HYPOMYELINATING, 2; HLD2",
    "OMIM:608807": "MUSCULAR DYSTROPHY, LIMB-GIRDLE, AUTOSOMAL RECESSIVE 10; LGMDR10",
    "OMIM:609033": "POSTERIOR COLUMN ATAXIA WITH RETINITIS PIGMENTOSA; AXPC1",
    "OMIM:609056": "SALT AND PEPPER DEVELOPMENTAL REGRESSION SYNDROME; SPDRS",
    "OMIM:609195": "SPASTIC PARAPLEGIA 26, AUTOSOMAL RECESSIVE; SPG26",
    "OMIM:609200": "MYOPATHY, MYOFIBRILLAR, 3; MFM3",
    "OMIM:609260": "CHARCOT-MARIE-TOOTH DISEASE, AXONAL, AUTOSOMAL DOMINANT, TYPE 2A2A; CMT2A2A",
    "OMIM:609270": "SPINOCEREBELLAR ATAXIA, AUTOSOMAL RECESSIVE 7; SCAR7",
    "OMIM:609304": "DEVELOPMENTAL AND EPILEPTIC ENCEPHALOPATHY 3; DEE3",
    "OMIM:609524": "MYOPATHY, MYOFIBRILLAR, 5; MFM5",
    "OMIM:609560": "MITOCHONDRIAL DNA DEPLETION SYNDROME 2 (MYOPATHIC TYPE); MTDPS2",
    "OMIM:609923": "RETINITIS PIGMENTOSA 31; RP31",
    "OMIM:610019": "CATARACT 18; CTRCT18",
    "OMIM:610031": "CORTICAL DYSPLASIA, COMPLEX, WITH OTHER BRAIN MALFORMATIONS 7; CDCBM7",
    "OMIM:610127": "CEROID LIPOFUSCINOSIS, NEURONAL, 10; CLN10",
    "OMIM:610181": "AICARDI-GOUTIERES SYNDROME 2; AGS2",
    "OMIM:610188": "JOUBERT SYNDROME 5; JBTS5",
    "OMIM:610217": "NEURODEGENERATION WITH BRAIN IRON ACCUMULATION 2B; NBIA2B",
    "OMIM:610329": "AICARDI-GOUTIERES SYNDROME 3; AGS3",
    "OMIM:610333": "AICARDI-GOUTIERES SYNDROME 4; AGS4",
    "OMIM:610359": "RETINITIS PIGMENTOSA 33; RP33",
    "OMIM:610443": "KOOLEN-DE VRIES SYNDROME; KDVS",
    "OMIM:610444": "NIGHT BLINDNESS, CONGENITAL STATIONARY, AUTOSOMAL DOMINANT 3; CSNBAD3",
    "OMIM:610489": "PIGMENTED NODULAR ADRENOCORTICAL DISEASE, PRIMARY, 1; PPNAD1",
    "OMIM:610532": "LEUKODYSTROPHY, HYPOMYELINATING, 5; HLD5",
    "OMIM:610688": "JOUBERT SYNDROME 6; JBTS6",
    "OMIM:610725": "NEPHROTIC SYNDROME, TYPE 3; NPHS3",
    "OMIM:610951": "CEROID LIPOFUSCINOSIS, NEURONAL, 7; CLN7",
    "OMIM:611091": "MENTAL RETARDATION, AUTOSOMAL RECESSIVE 5; MRT5",
    "OMIM:611302": "SPASTIC ATAXIA 2, AUTOSOMAL RECESSIVE; SPAX2",
    "OMIM:611307": "MUSCULAR DYSTROPHY, LIMB-GIRDLE, AUTOSOMAL RECESSIVE 12; LGMDR12",
    "OMIM:611560": "JOUBERT SYNDROME 7; JBTS7",
    "OMIM:611603": "LISSENCEPHALY 3; LIS3",
    "OMIM:611705": "SALIH MYOPATHY; SALMY",
    "OMIM:611721": "COMBINED SAPOSIN DEFICIENCY",
    "OMIM:611726": "EPILEPSY, PROGRESSIVE MYOCLONIC, 3, WITH OR WITHOUT INTRACELLULAR INCLUSIONS; EPM3",
    "OMIM:611755": "LEBER CONGENITAL AMAUROSIS 10; LCA10",
    "OMIM:612015": "CONGENITAL DISORDER OF GLYCOSYLATION, TYPE In; CDG1N",
    "OMIM:612095": "RETINITIS PIGMENTOSA 41; RP41",
    "OMIM:612126": "GLUT1 DEFICIENCY SYNDROME 2; GLUT1DS2",
    "OMIM:612285": "JOUBERT SYNDROME 9; JBTS9",
    "OMIM:612337": "MENTAL RETARDATION, AUTOSOMAL DOMINANT 22; MRD22",
    "OMIM:612370": "HYPOGONADOTROPIC HYPOGONADISM 5 WITH OR WITHOUT ANOSMIA; HH5",
    "OMIM:612438": "LEUKODYSTROPHY, HYPOMYELINATING, 6; HLD6",
    "OMIM:612528": "DIAMOND-BLACKFAN ANEMIA 5; DBA5",
    "OMIM:612621": "MENTAL RETARDATION, AUTOSOMAL DOMINANT 5; MRD5",
    "OMIM:612712": "LEBER CONGENITAL AMAUROSIS 13; LCA13",
    "OMIM:612716": "DYSTONIA, DOPA-RESPONSIVE, DUE TO SEPIAPTERIN REDUCTASE DEFICIENCY",
    "OMIM:612780": "SEIZURES, SENSORINEURAL DEAFNESS, ATAXIA, MENTAL RETARDATION, AND ELECTROLYTE IMBALANCE; SESAMES",
    "OMIM:612936": "SPASTIC PARAPLEGIA 50, AUTOSOMAL RECESSIVE; SPG50",
    "OMIM:612943": "RETINITIS PIGMENTOSA 42; RP42",
    "OMIM:612951": "LEUKOENCEPHALOPATHY, CYSTIC, WITHOUT MEGALENCEPHALY",
    "OMIM:613065": "LEUKEMIA, ACUTE LYMPHOBLASTIC; ALL",
    "OMIM:613091": "SHORT-RIB THORACIC DYSPLASIA 3 WITH OR WITHOUT POLYDACTYLY; SRTD3",
    "OMIM:613155": "MUSCULAR DYSTROPHY-DYSTROGLYCANOPATHY (CONGENITAL WITH IMPAIRED INTELLECTUAL IMPAIRMENT), TYPE B, 1; MDDGB1",
    "OMIM:613156": "MUSCULAR DYSTROPHY-DYSTROGLYCANOPATHY (CONGENITAL WITH IMPAIRED INTELLECTUAL DEVELOPMENT), TYPE B, 2; MDDGB2",
    "OMIM:613158": "MUSCULAR DYSTROPHY-DYSTROGLYCANOPATHY (LIMB-GIRDLE), TYPE C, 2; MDDGC2",
    "OMIM:613192": "MENTAL RETARDATION, AUTOSOMAL RECESSIVE 13; MRT13",
    "OMIM:613205": "MUSCULAR DYSTROPHY, CONGENITAL, LMNA-RELATED",
    "OMIM:613216": "NIGHT BLINDNESS, CONGENITAL STATIONARY, TYPE 1C; CSNB1C",
    "OMIM:613217": "DIARRHEA 5, WITH TUFTING ENTEROPATHY, CONGENITAL; DIAR5",
    "OMIM:613402": "MICROCEPHALY, SEIZURES, AND DEVELOPMENTAL DELAY; MCSZ",
    "OMIM:613428": "RETINITIS PIGMENTOSA 54; RP54",
    "OMIM:613454": "RETT SYNDROME, CONGENITAL VARIANT",
    "OMIM:613550": "NEPHRONOPHTHISIS 11; NPHP11",
    "OMIM:613561": "MYOPATHY, LACTIC ACIDOSIS, AND SIDEROBLASTIC ANEMIA 2; MLASA2",
    "OMIM:613587": "OCCULT MACULAR DYSTROPHY; OCMD",
    "OMIM:613612": "CONGENITAL DISORDER OF GLYCOSYLATION, TYPE IIi; CDG2I",
    "OMIM:613660": "CONE-ROD DYSTROPHY 15; CORD15",
    "OMIM:613668": "MICROCEPHALY, POSTNATAL PROGRESSIVE, WITH SEIZURES AND BRAIN ATROPHY",
    "OMIM:613680": "BEAULIEU-BOYCOTT-INNES SYNDROME; BBIS",
    "OMIM:613684": "RUBINSTEIN-TAYBI SYNDROME 2; RSTS2",
    "OMIM:613690": "CARDIOMYOPATHY, FAMILIAL HYPERTROPHIC, 7; CMH7",
    "OMIM:613706": "NOONAN SYNDROME 7; NS7",
    "OMIM:613722": "DEVELOPMENTAL AND EPILEPTIC ENCEPHALOPATHY 12; DEE12",
    "OMIM:613731": "RETINITIS PIGMENTOSA 4; RP4",
    "OMIM:613744": "SPASTIC PARAPLEGIA 51, AUTOSOMAL RECESSIVE; SPG51",
    "OMIM:613750": "RETINITIS PIGMENTOSA 27; RP27",
    "OMIM:613756": "RETINITIS PIGMENTOSA 49; RP49",
    "OMIM:613767": "RETINITIS PIGMENTOSA 45; RP45",
    "OMIM:613794": "RETINITIS PIGMENTOSA 20; RP20",
    "OMIM:613809": "RETINITIS PIGMENTOSA 39; RP39",
    "OMIM:613810": "RETINITIS PIGMENTOSA 43; RP43",
    "OMIM:613811": "PONTOCEREBELLAR HYPOPLASIA, TYPE 2D; PCH2D",
    "OMIM:613820": "NEPHRONOPHTHISIS 12; NPHP12",
    "OMIM:613835": "LEBER CONGENITAL AMAUROSIS 8; LCA8",
    "OMIM:613861": "RETINITIS PIGMENTOSA 59; RP59",
    "OMIM:613862": "RETINITIS PIGMENTOSA 38; RP38",
    "OMIM:614034": "HEME OXYGENASE 1 DEFICIENCY; HMOX1D",
    "OMIM:614065": "MYOPATHY, DISTAL, 4; MPD4",
    "OMIM:614173": "JOUBERT SYNDROME 13; JBTS13",
    "OMIM:614180": "RETINITIS PIGMENTOSA 61; RP61",
    "OMIM:614181": "RETINITIS PIGMENTOSA 62; RP62",
    "OMIM:614199": "NEPHROTIC SYNDROME, TYPE 5, WITH OR WITHOUT OCULAR ABNORMALITIES; NPHS5",
    "OMIM:614202": "RAFIQ SYNDROME; RAFQS",
    "OMIM:614225": "WARBURG MICRO SYNDROME 2; WARBM2",
    "OMIM:614254": "NEURODEVELOPMENTAL DISORDER WITH OR WITHOUT HYPERKINETIC MOVEMENTS AND SEIZURES, AUTOSOMAL DOMINANT; NDHMSD",
    "OMIM:614255": "NESCAV SYNDROME; NESCAVS",
    "OMIM:614299": "MULTIPLE MITOCHONDRIAL DYSFUNCTIONS SYNDROME 2 WITH HYPERGLYCINEMIA; MMDS2",
    "OMIM:614381": "LEUKODYSTROPHY, HYPOMYELINATING, 8, WITH OR WITHOUT OLIGODONTIA AND/OR HYPOGONADOTROPIC HYPOGONADISM; HLD8",
    "OMIM:614424": "JOUBERT SYNDROME 14; JBTS14",
    "OMIM:614500": "CONE-ROD DYSTROPHY 16; CORD16",
    "OMIM:614557": "EHLERS-DANLOS SYNDROME, KYPHOSCOLIOTIC TYPE, 2; EDSKSCL2",
    "OMIM:614607": "COFFIN-SIRIS SYNDROME 2; CSS2",
    "OMIM:614615": "JOUBERT SYNDROME 17; JBTS17",
    "OMIM:614669": "AURICULOCONDYLAR SYNDROME 2; ARCND2",
    "OMIM:614678": "PONTOCEREBELLAR HYPOPLASIA, TYPE 1B; PCH1B",
    "OMIM:614744": "FACIAL PARESIS, HEREDITARY CONGENITAL, 3; HCFP3",
    "OMIM:614753": "SOTOS SYNDROME 2; SOTOS2",
    "OMIM:614831": "SPINOCEREBELLAR ATAXIA, AUTOSOMAL RECESSIVE 13; SCAR13",
    "OMIM:614833": "MICROCEPHALY, SHORT STATURE, AND POLYMICROGYRIA WITH OR WITHOUT SEIZURES; MSSP",
    "OMIM:614876": "PEROXISOME BIOGENESIS DISORDER 8A (ZELLWEGER); PBD8A",
    "OMIM:614877": "PEROXISOME BIOGENESIS DISORDER 8B; PBD8B",
    "OMIM:614932": "COMBINED OXIDATIVE PHOSPHORYLATION DEFICIENCY 13; COXPD13",
    "OMIM:614945": "DEAFNESS, AUTOSOMAL RECESSIVE 18B; DFNB18B",
    "OMIM:615030": "SPASTIC PARAPLEGIA 56, AUTOSOMAL RECESSIVE; SPG56",
    "OMIM:615031": "SPASTIC PARAPLEGIA 49, AUTOSOMAL RECESSIVE; SPG49",
    "OMIM:615065": "ARTHROGRYPOSIS, DISTAL, TYPE 5D; DA5D",
    "OMIM:615071": "ALAZAMI SYNDROME; ALAZS",
    "OMIM:615095": "MICROCEPHALY 10, PRIMARY, AUTOSOMAL RECESSIVE; MCPH10",
    "OMIM:615181": "MUSCULAR DYSTROPHY-DYSTROGLYCANOPATHY (CONGENITAL WITH BRAIN AND EYE ANOMALIES), TYPE A, 11; MDDGA11",
    "OMIM:615190": "DYSKERATOSIS CONGENITA, AUTOSOMAL RECESSIVE 5; DKCB5",
    "OMIM:615191": "LISSENCEPHALY 5; LIS5",
    "OMIM:615268": "CEREBELLAR ATAXIA, MENTAL RETARDATION, AND DYSEQUILIBRIUM SYNDROME 4; CAMRQ4",
    "OMIM:615286": "MENTAL RETARDATION, AUTOSOMAL RECESSIVE 36; MRT36",
    "OMIM:615352": "MUSCULAR DYSTROPHY-DYSTROGLYCANOPATHY (LIMB-GIRDLE), TYPE C, 14; MDDGC14",
    "OMIM:615411": "CORTICAL DYSPLASIA, COMPLEX, WITH OTHER BRAIN MALFORMATIONS 3; CDCBM3",
    "OMIM:615412": "CORTICAL DYSPLASIA, COMPLEX, WITH OTHER BRAIN MALFORMATIONS 4; CDCBM4",
    "OMIM:615419": "HYPOTONIA, INFANTILE, WITH PSYCHOMOTOR RETARDATION AND CHARACTERISTIC FACIES 1; IHPRF1",
    "OMIM:615471": "MITOCHONDRIAL DNA DEPLETION SYNDROME 13 (ENCEPHALOMYOPATHIC TYPE); MTDPS13",
    "OMIM:615476": "DEVELOPMENTAL AND EPILEPTIC ENCEPHALOPATHY 18; DEE18",
    "OMIM:615485": "BAINBRIDGE-ROPERS SYNDROME; BRPS",
    "OMIM:615502": "MENTAL RETARDATION, AUTOSOMAL DOMINANT 21; MRD21",
    "OMIM:615546": "VAN MALDERGEM SYNDROME 2; VMLDS2",
    "OMIM:615574": "ASPARAGINE SYNTHETASE DEFICIENCY; ASNSD",
    "OMIM:615681": "SPASTIC PARAPLEGIA 62, AUTOSOMAL RECESSIVE; SPG62",
    "OMIM:615715": "BONE MARROW FAILURE SYNDROME 2; BMFS2",
    "OMIM:615716": "HYPERPHOSPHATASIA WITH MENTAL RETARDATION SYNDROME 4; HPMRS4",
    "OMIM:615760": "MICROCEPHALY, PROGRESSIVE, WITH SEIZURES AND CEREBRAL AND CEREBELLAR ATROPHY; MSCCA",
    "OMIM:615763": "CORTICAL DYSPLASIA, COMPLEX, WITH OTHER BRAIN MALFORMATIONS 5; CDCBM5",
    "OMIM:615809": "PONTOCEREBELLAR HYPOPLASIA, TYPE 9; PCH9",
    "OMIM:615829": "XIA-GIBBS SYNDROME; XIGIS",
    "OMIM:615834": "MENTAL RETARDATION, AUTOSOMAL DOMINANT 26; MRD26",
    "OMIM:615846": "AICARDI-GOUTIERES SYNDROME 7; AGS7",
    "OMIM:615866": "COFFIN-SIRIS SYNDROME 9; CSS9",
    "OMIM:615909": "DIAMOND-BLACKFAN ANEMIA 13; DBA13",
    "OMIM:615948": "OROFACIODIGITAL SYNDROME XIV; OFD14",
    "OMIM:615959": "MYOPATHY, CENTRONUCLEAR, 5; CNM5",
    "OMIM:615960": "PORETTI-BOLTSHAUSER SYNDROME; PTBHS",
    "OMIM:615973": "CONE-ROD DYSTROPHY 20; CORD20",
    "OMIM:615981": "BARDET-BIEDL SYNDROME 2; BBS2",
    "OMIM:615982": "BARDET-BIEDL SYNDROME 4; BBS4",
    "OMIM:615983": "BARDET-BIEDL SYNDROME 5; BBS5",
    "OMIM:615993": "BARDET-BIEDL SYNDROME 16; BBS16",
    "OMIM:616051": "MICROCEPHALY 13, PRIMARY, AUTOSOMAL RECESSIVE; MCPH13",
    "OMIM:616056": "DEVELOPMENTAL AND EPILEPTIC ENCEPHALOPATHY 26; DEE26",
    "OMIM:616081": "PONTOCEREBELLAR HYPOPLASIA, TYPE 1C; PCH1C",
    "OMIM:616094": "MUSCULAR DYSTROPHY-DYSTROGLYCANOPATHY (LIMB-GIRDLE), TYPE C, 12; MDDGC12",
    "OMIM:616127": "SPINOCEREBELLAR ATAXIA, AUTOSOMAL RECESSIVE 17; SCAR17",
    "OMIM:616140": "LEUKODYSTROPHY, HYPOMYELINATING, 9; HLD9",
    "OMIM:616155": "CHARCOT-MARIE-TOOTH DISEASE, AXONAL, TYPE 2S; CMT2S",
    "OMIM:616171": "MICROCEPHALY AND CHORIORETINOPATHY, AUTOSOMAL RECESSIVE, 2; MCCRP2",
    "OMIM:616199": "POLYGLUCOSAN BODY MYOPATHY 2; PGBM2",
    "OMIM:616211": "DEVELOPMENTAL AND EPILEPTIC ENCEPHALOPATHY 28; DEE28",
    "OMIM:616263": "NEUROLOGIC, ENDOCRINE, AND PANCREATIC DISEASE, MULTISYSTEM, INFANTILE-ONSET; IMNEPD",
    "OMIM:616266": "CONGENITAL CONTRACTURES OF THE LIMBS AND FACE, HYPOTONIA, AND DEVELOPMENTAL DELAY; CLIFAHDD",
    "OMIM:616271": "3-METHYLGLUTACONIC ACIDURIA, TYPE VII; MGCA7",
    "OMIM:616277": "MITOCHONDRIAL SHORT-CHAIN ENOYL-CoA HYDRATASE 1 DEFICIENCY; ECHS1D",
    "OMIM:616281": "NEURODEVELOPMENTAL DISORDER WITH SPASTIC PARAPLEGIA AND MICROCEPHALY; NEDSPM",
    "OMIM:616321": "MYASTHENIC SYNDROME, CONGENITAL, 3A, SLOW-CHANNEL; CMS3A",
    "OMIM:616354": "SPINOCEREBELLAR ATAXIA, AUTOSOMAL RECESSIVE 20; SCAR20",
    "OMIM:616394": "RETINITIS PIGMENTOSA 71; RP71",
    "OMIM:616420": "LEUKODYSTROPHY, HYPOMYELINATING, 10; HLD10",
    "OMIM:616421": "MYOCLONIC-ATONIC EPILEPSY; MAE",
    "OMIM:616471": "BETHLEM MYOPATHY 2; BTHLM2",
    "OMIM:616490": "JOUBERT SYNDROME 23; JBTS23",
    "OMIM:616531": "POLYMICROGYRIA, PERISYLVIAN, WITH CEREBELLAR HYPOPLASIA AND ARTHROGRYPOSIS; PMGYCHA",
    "OMIM:616544": "RETINITIS PIGMENTOSA 73; RP73",
    "OMIM:616562": "RETINITIS PIGMENTOSA 74; RP74",
    "OMIM:616564": "NOONAN SYNDROME 10; NS10",
    "OMIM:616632": "SEIZURES, CORTICAL BLINDNESS, AND MICROCEPHALY SYNDROME; SCBMS",
    "OMIM:616647": "DEVELOPMENTAL AND EPILEPTIC ENCEPHALOPATHY 35; DEE35",
    "OMIM:616649": "SPHEROCYTOSIS, TYPE 2; SPH2",
    "OMIM:616651": "ROIFMAN SYNDROME; RFMN",
    "OMIM:616654": "JOUBERT SYNDROME 24; JBTS24",
    "OMIM:616657": "SPASTIC TETRAPLEGIA, THIN CORPUS CALLOSUM, AND PROGRESSIVE MICROCEPHALY; SPATCCM",
    "OMIM:616668": "CHARCOT-MARIE-TOOTH DISEASE, AXONAL, TYPE 2X; CMT2X",
    "OMIM:616683": "LEUKODYSTROPHY, HYPOMYELINATING, 12; HLD12",
    "OMIM:616721": "CONGENITAL DISORDER OF GLYCOSYLATION, TYPE IIn; CDG2N",
    "OMIM:616732": "OPTIC ATROPHY 10 WITH OR WITHOUT ATAXIA, MENTAL RETARDATION, AND SEIZURES; OPA10",
    "OMIM:616738": "RADIOULNAR SYNOSTOSIS WITH AMEGAKARYOCYTIC THROMBOCYTOPENIA 2; RUSAT2",
    "OMIM:616740": "IMMUNODEFICIENCY 46; IMD46",
    "OMIM:616789": "MENTAL RETARDATION AND DISTINCTIVE FACIAL FEATURES WITH OR WITHOUT CARDIAC DEFECTS; MRFACD",
    "OMIM:616801": "HYPOTONIA, INFANTILE, WITH PSYCHOMOTOR RETARDATION AND CHARACTERISTIC FACIES 2; IHPRF2",
    "OMIM:616875": "CEREBELLAR ATROPHY, VISUAL IMPAIRMENT, AND PSYCHOMOTOR RETARDATION; CAVIPMR",
    "OMIM:616892": "NEPHROTIC SYNDROME, TYPE 12; NPHS12",
    "OMIM:616900": "HYPOTONIA, INFANTILE, WITH PSYCHOMOTOR RETARDATION AND CHARACTERISTIC FACIES 3; IHPRF3",
    "OMIM:616973": "MENTAL RETARDATION, AUTOSOMAL DOMINANT 42; MRD42",
    "OMIM:617011": "MACROCEPHALY, DYSMORPHIC FACIES, AND PSYCHOMOTOR RETARDATION; MDFPMR",
    "OMIM:617013": "HYPERMANGANESEMIA WITH DYSTONIA 2; HMNDYT2",
    "OMIM:617047": "CARDIOMYOPATHY, FAMILIAL HYPERTROPHIC, 26; CMH26",
    "OMIM:617052": "BONE MARROW FAILURE SYNDROME 3; BMFS3",
    "OMIM:617053": "MIRAGE SYNDROME; MIRAGE",
    "OMIM:617093": "GROWTH RETARDATION, IMPAIRED INTELLECTUAL DEVELOPMENT, HYPOTONIA, AND HEPATOPATHY; GRIDHH",
    "OMIM:617106": "DEVELOPMENTAL AND EPILEPTIC ENCEPHALOPATHY 42; DEE42",
    "OMIM:617120": "JOUBERT SYNDROME 27; JBTS27",
    "OMIM:617121": "JOUBERT SYNDROME 28; JBTS28",
    "OMIM:617123": "RETINITIS PIGMENTOSA 76; RP76",
    "OMIM:617137": "FRONTOMETAPHYSEAL DYSPLASIA 2; FMD2",
    "OMIM:617146": "ARTHROGRYPOSIS, DISTAL, WITH IMPAIRED PROPRIOCEPTION AND TOUCH; DAIPT",
    "OMIM:617183": "HAREL-YOON SYNDROME; HAYOS",
    "OMIM:617193": "ENCEPHALOPATHY, PROGRESSIVE, EARLY-ONSET, WITH BRAIN ATROPHY AND THIN CORPUS CALLOSUM; PEBAT",
    "OMIM:617222": "SUDDEN CARDIAC FAILURE, INFANTILE; SCFI",
    "OMIM:617276": "DEVELOPMENTAL AND EPILEPTIC ENCEPHALOPATHY 48; DEE48",
    "OMIM:617330": "HYPOTONIA, ATAXIA, AND DELAYED DEVELOPMENT SYNDROME; HADDS",
    "OMIM:617350": "DEVELOPMENTAL AND EPILEPTIC ENCEPHALOPATHY 52; DEE52",
    "OMIM:617397": "PSEUDO-TORCH SYNDROME 2; PTORCH2",
    "OMIM:617433": "RETINITIS PIGMENTOSA 78; RP78",
    "OMIM:617443": "BLEEDING DISORDER, PLATELET-TYPE, 21; BDPLT21",
    "OMIM:617460": "RETINITIS PIGMENTOSA 79; RP79",
    "OMIM:617481": "NEURODEVELOPMENTAL DISORDER WITH MICROCEPHALY, HYPOTONIA, AND VARIABLE BRAIN ANOMALIES; NMIHBA",
    "OMIM:617537": "RAHMAN SYNDROME; RMNS",
    "OMIM:617547": "RETINAL DYSTROPHY WITH OR WITHOUT MACULAR STAPHYLOMA; RDMS",
    "OMIM:617622": "JOUBERT SYNDROME 30; JBTS30",
    "OMIM:617675": "MYOPATHY, MITOCHONDRIAL, AND ATAXIA; MMYAT",
    "OMIM:617681": "BLEPHAROCHEILODONTIC SYNDROME 2; BCDS2",
    "OMIM:617694": "AL KAISSI SYNDROME; ALKAS",
    "OMIM:617781": "RETINITIS PIGMENTOSA 80; RP80",
    "OMIM:617805": "RENAL HYPODYSPLASIA/APLASIA 3; RHDA3",
    "OMIM:617807": "NEURODEVELOPMENTAL DISORDER WITH ATAXIC GAIT, ABSENT SPEECH, AND DECREASED CORTICAL WHITE MATTER; NDAGSCW",
    "OMIM:617829": "EPILEPTIC ENCEPHALOPATHY, INFANTILE OR EARLY CHILDHOOD, 2; IECEE2",
    "OMIM:617873": "COMBINED OXIDATIVE PHOSPHORYLATION DEFICIENCY 35; COXPD35",
    "OMIM:617948": "ELLIPTOCYTOSIS 3; EL3",
    "OMIM:617954": "MULTIPLE MITOCHONDRIAL DYSFUNCTIONS SYNDROME 6; MMDS6",
    "OMIM:618056": "NEURODEVELOPMENTAL DISORDER WITH CEREBELLAR ATROPHY AND WITH OR WITHOUT SEIZURES; NEDCAS",
    "OMIM:618088": "NEURODEVELOPMENTAL DISORDER WITH REGRESSION, ABNORMAL MOVEMENTS, LOSS OF SPEECH, AND SEIZURES; NEDAMSS",
    "OMIM:618129": "MUSCULAR DYSTROPHY, LIMB-GIRDLE, AUTOSOMAL DOMINANT 4; LGMDD4",
    "OMIM:618138": "MUSCULAR DYSTROPHY, LIMB-GIRDLE, AUTOSOMAL RECESSIVE 23; LGMDR23",
    "OMIM:618174": "CORTICAL DYSPLASIA, COMPLEX, WITH OTHER BRAIN MALFORMATIONS 9; CDCBM9",
    "OMIM:618225": "MITOCHONDRIAL COMPLEX I DEFICIENCY, NUCLEAR TYPE 4; MC1DN4",
    "OMIM:618273": "MEGA-CORPUS-CALLOSUM SYNDROME WITH CEREBELLAR HYPOPLASIA AND CORTICAL MALFORMATIONS; MCCCHCM",
    "OMIM:618316": "INTELLECTUAL DEVELOPMENTAL DISORDER WITH CARDIAC DEFECTS AND DYSMORPHIC FACIES; IDDCDF",
    "OMIM:618325": "LISSENCEPHALY 9 WITH COMPLEX BRAINSTEM MALFORMATION; LIS9",
    "OMIM:618404": "LEUKODYSTROPHY, HYPOMYELINATING, 18; HLD18",
    "OMIM:618512": "O'DONNELL-LURIA-RODAN SYNDROME; ODLURO"
}

disease_system = {
    "DOID": "http://purl.obolibrary.org/obo/doid.owl",
    "HP": "http://purl.obolibrary.org/obo/hp.owl",
    "OMIM": "http://purl.obolibrary.org/obo/omim.owl"
}



"""Represent fhir entity."""
from anvil.transformers.fhir import CANONICAL, join, make_identifier

# specimen_type QUESTION: https://www.hl7.org/fhir/v2/0487/index.html
# specimen_type = {
#     constants.SPECIMEN.COMPOSITION.BLOOD: {
#         "system": "http://terminology.hl7.org/CodeSystem/v2-0487",
#         "code": "BLD",
#         "display": "Whole blood",
#     },
#     constants.SPECIMEN.COMPOSITION.SALIVA: {
#         "system": "http://terminology.hl7.org/CodeSystem/v2-0487",
#         "code": "SAL",
#         "display": "Saliva",
#     },
#     constants.SPECIMEN.COMPOSITION.TISSUE: {
#         "system": "http://terminology.hl7.org/CodeSystem/v2-0487",
#         "code": "TISS",
#         "display": "Tissue",
#     },
# }


class Specimen:
    """Create fhir entity."""

    class_name = "specimen"
    resource_type = "Specimen"

    @staticmethod
    def identifier(specimen):
        """Create identifier."""
        return make_identifier(specimen.workspace_name, specimen.id)

    @staticmethod
    def build_entity(specimen, subject):
        """Create fhir entity."""
        study_id = specimen.workspace_name
        # study_id_slug = make_identifier(study_id)
        sample_id = specimen.id
        sample_id_slug = make_identifier(sample_id)
        event_age_days = None
        concentration_mg_per_ml = None
        composition = None
        volume_ul = None
        subject_id_slug = make_identifier('P', subject.id)

        entity = {
            "resourceType": Specimen.resource_type,
            "id": sample_id_slug,
            "meta": {
                "profile": [
                    "http://hl7.org/fhir/StructureDefinition/Specimen"
                ]
            },
            "identifier": [
                {
                    "system": f"https://anvil.terra.bio/#workspaces/anvil-datastorage/{study_id}",
                    "value": sample_id,
                },
                {
                    "system": "urn:ncpi:unique-string",
                    "value": join(Specimen.resource_type, study_id, sample_id),
                },
            ],
            "subject": {
                "reference": f"Patient/{subject_id_slug}"
            },
        }

        # event_age_days: QUESTION extension ?
        if event_age_days:
            entity.setdefault("extension", []).append(
                {
                    "url": f"{CANONICAL}/StructureDefinition/age-at-event",
                    "valueAge": {
                        "value": int(event_age_days),
                        "unit": "d",
                        "system": "http://unitsofmeasure.org",
                        "code": "days",
                    },
                }
            )

        # concentration_mg_per_ml: QUESTION extension ?
        if concentration_mg_per_ml:
            entity.setdefault("extension", []).append(
                {
                    "url": f"{CANONICAL}/StructureDefinition/concentration",
                    "valueQuantity": {
                        "value": float(concentration_mg_per_ml),
                        "unit": "mg/mL",
                    },
                }
            )

        # composition: QUESTION extension ?
        if composition:
            entity["type"] = {
                "coding": "TODO",  # [specimen_type[composition]],
                "text": composition,
            }

        # volume_ul: QUESTION extension ?
        if volume_ul:
            entity.setdefault("collection", {})["quantity"] = {
                "unit": "uL",
                "value": float(volume_ul),
            }

        return entity



"""Render workspaces into a sqlite graph."""
import sqlite3
import json
import logging


class WorkspaceGraph:
    """Represent workspace as in a graph."""

    def __init__(self, path):
        """Initialize db."""
        self.path = path
        assert self.path, "Path must be set"
        self._conn = sqlite3.connect(self.path)
        cur = self._conn.cursor()
        cur.executescript("""
        CREATE TABLE IF NOT EXISTS vertices (
            key text PRIMARY KEY,
            submitter_id text,
            name text,
            json text NOT NULL
        );
        CREATE TABLE IF NOT EXISTS edges (
            src text,
            dst text,
            src_name text,
            dst_name text
        );
        """)
        self._conn.commit()

    def put(self, key, submitter_id, name, data, cur):
        """Save an item."""
        if hasattr(data, 'attributes'):
            _json = json.dumps(data.attributes)
        else:
            _json = json.dumps(data)
        cur.execute("REPLACE into vertices values (?, ?, ?, ?);", (key, submitter_id, name, _json))
        # self._logger.debug(f"put {key}")

    def get(self, key):
        """Retrieve an item."""
        cur = self._conn.cursor()
        data = cur.execute("SELECT json, key, submitter_id, name FROM vertices where key=?", (key,)).fetchone()
        if data:
            v = json.loads(data[0])
            v['_key'] = data[1]
            v['_submitter_id'] = data[2]
            v['_name'] = data[3]
            src_edges = [(d[0], d[1]) for d in cur.execute("SELECT src, src_name FROM edges where dst=?", (key,)).fetchall()]
            edges = {}
            for src in src_edges:
                if src[1] not in edges:
                    edges[src[1]] = []
                for d in cur.execute("SELECT json, key, name, submitter_id FROM vertices where key=?", (src[0],)).fetchall():
                    e = json.loads(d[0])
                    e['_key'] = d[1]
                    e['_name'] = d[2]
                    e['_submitter_id'] = d[3]
                    edges[src[1]].append(e)

            return {'vertex': v, 'edges': edges}
        assert False, f"NOT FOUND {key}"

    def get_by_name(self, name=None):
        """Retrieve all items with name."""
        cur = self._conn.cursor()
        data = cur.execute("SELECT key FROM vertices where name=?", (name,)).fetchall()
        return [self.get(d[0]) for d in data]

    def put_edge(self, src, dst, src_name, dst_name, cur):
        """Save edge."""
        cur.execute("REPLACE into edges values (?, ?, ?, ?);", (src, dst, src_name, dst_name))

    def save(self, workspace):
        """Save workspace to graph."""
        cur = self._conn.cursor()
        logging.getLogger(__name__).info(f'Loading {workspace.name}')
        self.put(workspace.id, workspace.name, 'workspace', workspace, cur)
        for subject in workspace.subjects:
            self.put(subject.id, subject.id, 'subject', subject, cur)
            self.put_edge(subject.id, workspace.id, 'subject', 'workspace', cur)
            for sample in subject.samples:
                self.put(sample.id, sample.id, 'sample', sample, cur)
                self.put_edge(sample.id, subject.id, 'sample', 'subject', cur)
                for blob_id, blob in sample.blobs.items():
                    self.put(blob_id, blob_id, 'blob', blob, cur)
                    self.put_edge(blob_id, sample.id, 'blob', 'sample', cur)
                    if 'ga4gh_drs_uri' in blob:
                        drs = {'uri': blob['ga4gh_drs_uri']}
                        self.put(drs['uri'], sample.id, 'drs', drs, cur)
                        self.put_edge(drs['uri'], sample.id, 'drs', 'sample', cur)

        self._conn.commit()

    def index(self):
        """Index the vertices and edges."""
        logging.getLogger(__name__).info('Indexing')
        cur = self._conn.cursor()
        cur.executescript("""
        CREATE  INDEX IF NOT EXISTS vertices_submitter_id ON vertices(submitter_id);
        CREATE UNIQUE INDEX IF NOT EXISTS edges_src_dst ON edges(src, dst, src_name, dst_name);
        CREATE  INDEX IF NOT EXISTS edges_dst ON edges(dst);
        """)
        self._conn.commit()



"""Extract AnVIL Data Ingestion Tracker spreadsheet."""

import gspread
from oauth2client.service_account import ServiceAccountCredentials
from collections import OrderedDict
import os

DEFAULT_OUTPUT_PATH = f"{os.path.dirname(__file__)}/../data/data_ingestion_tracker.json"


def download_projects(spreadsheet_key, json_keyfile_path):
    """Read projects from master spreadsheet."""
    # use creds to create a client to interact with the Google Drive API
    scope = ['https://spreadsheets.google.com/feeds', 'https://www.googleapis.com/auth/drive']
    credentials = ServiceAccountCredentials.from_json_keyfile_name(json_keyfile_path, scope)
    client = gspread.authorize(credentials)
    sheet = client.open_by_key(spreadsheet_key)
    for index, worksheet in enumerate(sheet.worksheets()):
        if 'Ingestion' in worksheet.title:
            break
    #  Extract all the rows
    list_of_lists = sheet.get_worksheet(index).get_all_values()
    list_of_lists.pop(0)  # junk
    # table headers
    keys = list_of_lists.pop(0)
    # re-write headers
    headers = OrderedDict({
        '': '',
        'Date AnVIL_Ingestion added as Reader to Workspace': 'date_AnVIL_ingestion_added_as_reader',
        'Date AnVIL_Ingestion added to Auth Domain': 'date_AnVIL_ingestion_added_to_auth_domain',
        'Date Auth Domain Added to Workspace as Reader': 'date_authorization_domain_added_to_workspace',
        'Date Authorization Domain Setup': 'date_authorization_domain_setup',
        'Date Consortium List Added to Auth Domain': 'date_consortium_added_to_auth_domain',
        'Date DUL/Consent Exists': 'date_dul_consent_exists',
        'Date Manifest File Created and Deposited to Box': 'date_manifest_file_deposited',
        'Date Terra Data Tables are Complete ': 'date_terra_data_tables_are_complete ',
        'Date Workspace Description Added': 'date_workspace_description_added',
        'Date dbGaP Linkage to Auth Domain': 'date_dbGaP_linkage_to_auth_domain',
        'Expected Final Number of Samples': 'expected_final_number_of_samples',
        'External Data Source = AnVIL in dbGaP': 'external_data_source',
        'FULL dbGaP Accession or other ACL\n(phs00000.v0.p0.c0)': 'dbGaP_accession',
        'Gen Data Transfer': 'gen_data_transfer',
        'Gen3 Program': 'gen3_program',
        'Gen3 Project': 'gen3_project',
        'INT Google Bucket Access Group': 'INT_google_bucket_access_group',
        'INT Service Account': 'INT_service_account',
        'Indexing Status': 'indexing_status',
        'Latest released version consent codes': 'latest_released_version_consent_codes',
        'Metric File Exists in Bucket': 'metric_file_exists_in_bucket',
        'Metrics Merged into Terra Data Table': 'metrics_merged_into_terra_data_table',
        'Minimal Metadata Transfer': 'minimal_metadata_transfer',
        'Model Build': 'model_build',
        'Model Definition': 'model_definition',
        'N Samples in Dataset Attributes': 'n_samples_in_dataset_attributes',
        'Name of Authorization Domain': 'name_of_authorization_domain',
        'PROD Google Bucket Access Group': 'prod_google_bucket_access_group',
        'PROD Service Account': 'prod_service_account',
        'Pheno Data Transfer': 'pheno_data_transfer',
        'Pheno QC': 'pheno_qc',
        'Pheno. Submission': 'pheno_submission',
        'Project': 'project',
        'Proposed Consent Code': 'proposed_consent_code',
        'Proposed Gen3 Project Name': 'proposed_gen3_project_name',
        'QC': 'QC',
        'QC Metadata FIle': 'qc_metadata_file',
        'Ready to Index': 'ready_to_index',
        'Relevant Consortium List': 'relevant_consortium_list',
        'Reprocessing Complete': 'reprocessing_complete',
        'Reprocessing Needed?': 'reprocessing_needed',
        'Run QA WDL': 'run_qa_wdl',
        'STAGING Google Bucket Access Group': 'staging_google_bucket_access_group',
        'Sequencing Center': 'sequencing_center',
        'Sequencing Strategy (WGS/WES/Array/Multiple)': 'sequencing_strategy',
        'Service Account': 'service_account',
        'Should be displayed on Portal': 'display_on_portal',
        'Summary Statistics Loaded to Portal?': 'summary_statistics_loaded_to_portal',
        'Terra Bucket': 'terra_bucket',
        'Workspace Link': 'workspace_link',
        'Workspace Name': 'workspace_name',
        'dbGaP Linking Auth Group': 'dbGaP Linking Auth Group',
        'dbGaP Release Status (Released, In Progress, Not Begun, Not Intending To)': 'dbGaP_release_status',
        'simple Accession\n(or "Pending")': 'simple_accession',
        'AnVIL_Discovery added to Workspace as Reader': 'AnVIL_Discovery_added_to_workspace',
        'BSB': 'BSB',
        'Consent Category and Research Use Limitations': 'consent_category',
        'Contact Email': 'contact_email',
        'Contact Person': 'contact_person',
        'Current Samples in Workspace (Data)': 'current_samples_in_workspace',
        'Data Tables Updates with Reprocessed Files': 'data_tables_updated_with_reprocessed_files',
        'Data Transformation': 'data_transformation',
        'Data Use Codes': 'data_use_codes',
        'Date AnVIL_Devs added to Auth Domain': 'date_anvil_devs_added_to_auth_domain'
    })

    keys = [headers.get(k, k) for k in keys]

    def normalize_value(v):
        """Fix the value."""
        if v == 'None':
            return None
        if v == '':
            return None
        if 'no' == str(v).lower():
            return False
        if 'yes' == str(v).lower():
            return True
        if str(v).isnumeric():
            return int(v)
        return v

    for lst in list_of_lists:
        project = {keys[i]: normalize_value(v) for i, v in enumerate(lst)}
        if project['workspace_name']:
            yield {keys[i]: normalize_value(v) for i, v in enumerate(lst)}



"""Represent fhir entity."""


from anvil.transformers.fhir import join, make_identifier


class Practitioner:
    """Create fhir entity."""

    class_name = "practitioner"
    resource_type = "Practitioner"

    @staticmethod
    def build_entity(workspace):
        """Create fhir entity."""
        study_id = workspace.id
        investigator_name = workspace.investigator
        if not investigator_name:
            return None

        workspace = workspace.attributes.workspace.attributes
        entity = {
            "resourceType": Practitioner.resource_type,
            "id": make_identifier(Practitioner.resource_type, investigator_name),
            "meta": {
                "profile": [
                    "http://hl7.org/fhir/StructureDefinition/Practitioner"
                ]
            },
            "identifier": [
                {
                    "system": "https://anvil.terra.bio/#workspaces/anvil-datastorage/",
                    "value": study_id,
                },
                {
                    "system": "urn:ncpi:unique-string",
                    "value": join(Practitioner.resource_type, investigator_name),
                },
            ],
            "name": [{"text": investigator_name}],
        }

        return entity



"""Reconcile and aggregate results."""

from anvil.dbgap.api import get_accession
from anvil.dbgap.api import get_study
from anvil.terra.reconciler import Reconciler
from collections import defaultdict
import logging

DEFAULT_NAMESPACE = 'anvil-datastorage'


def reconcile(name, user_project, namespace, workspace_regex, avro_path=None):
    """Run a reconciler on a set of workspaces, ."""
    logger = logging.getLogger('anvil.util.reconciler')
    reconciler = Reconciler(name, user_project, namespace, workspace_regex, avro_path)
    reconciler.save()
    reconciled_schemas = reconciler.reconcile_schemas()
    reconciled_schemas['name'] = name
    for view in reconciler.dashboard_views:
        accession = get_accession(namespace, view['project_id'])
        if accession:
            view['accession'] = accession
            study_tuple = get_study(accession)
            if not study_tuple:
                logger.warning(f"No study found {view['project_id']} accession: {accession}")
                view['problems'].append('missing_accession')
            else:
                qualified_accession, study = study_tuple
                view['qualified_accession'] = qualified_accession
                try:
                    view['dbgap_sample_count'] = len(study['DbGap']['Study']['SampleList']['Sample'])
                except Exception as e:
                    logger.warning(f"Study missing sample list {view['project_id']} accession: {accession} {e}")
                    view['dbgap_sample_count'] = 0
        yield view
    yield reconciled_schemas


def aggregate(namespace, user_project, consortium, avro_path=None):
    """Run a series of reconciliations."""
    def counts_factory():
        return {'expected_sample_count': 0, 'actual_sample_count': 0, 'problems': []}
    accessions = defaultdict(counts_factory)
    for name, workspace_regex in consortium:
        for view in reconcile(name, user_project, namespace, workspace_regex, avro_path):
            if 'qualified_accession' in view:
                accessions[view['qualified_accession']]['expected_sample_count'] = view['dbgap_sample_count']
                accessions[view['qualified_accession']]['actual_sample_count'] += [n for n in view['nodes'] if n['type'] == 'Samples'][0]['count']
            yield view
    for accession, counts in accessions.items():
        counts['workspace'] = None
        counts['qualified_accession'] = accession
        counts['project_id'] = accession
        counts['source'] = 'dbGAP'
        if counts['actual_sample_count'] != counts['expected_sample_count']:
            counts['problems'].append('dbgap_sample_count_mismatch')
        yield counts


def flatten(aggregations):
    """Render a dashboard data as key/value (flattened_aggregations, column_names)."""
    file_types = set()
    for p in aggregations:
        for f in p.get('files', []):
            file_types.add(f['type'])
    file_types = sorted(list(file_types))

    node_types = set()
    for p in aggregations:
        for n in p.get('nodes', []):
            node_types.add(n['type'])
    node_types = sorted(list(node_types))

    problem_types = set()
    for p in aggregations:
        for n in p.get('problems', []):
            problem_types.add(n)
    problem_types = sorted(list(problem_types))

    flattened = []
    for p in aggregations:
        flat = [p['source'], p['project_id'], p.get('qualified_accession', None)]

        file_sizes = [''] * len(file_types)
        for f in p.get('files', []):
            file_sizes[file_types.index(f['type'])] = f['size']
        flat.extend(file_sizes)
        flat.append(p.get('size', 0))

        node_counts = [''] * len(node_types)
        for n in p.get('nodes', []):
            node_counts[node_types.index(n['type'])] = n['count']
        flat.extend(node_counts)

        problems = [''] * len(problem_types)
        for n in p.get('problems', []):
            problems[problem_types.index(n)] = True
        flat.extend(problems)

        flattened.append(flat)

    return flattened, ['source', 'workspace', 'accession'] + file_types + ['size'] + node_types + problem_types



"""Cache items into sqlite."""
import os
import sqlite3
import json
import functools
import logging
from datetime import date, datetime, timedelta

CACHE_PATH = os.getenv('PYANVIL_CACHE_PATH', '/tmp/pyanvil-cache.sqlite')


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code."""
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError("Type %s not serializable" % type(obj))


class Cache():
    """Cache items in sqlite."""

    def __init__(self, path=CACHE_PATH, timeout=60 * 60 * 24):
        """Set up sqlite db."""
        logging.getLogger(__name__).debug(path)
        self._conn = sqlite3.connect(path)
        self._timeout = timeout
        cur = self._conn.cursor()
        cur.execute("""
        CREATE TABLE IF NOT EXISTS items (
            key text PRIMARY KEY,
            expiry TIMESTAMP,
            json text NOT NULL
        );""")
        self._conn.commit()
        # now = datetime.now().replace(microsecond=0).isoformat()
        # cur.execute("DELETE FROM items where expiry < ?", (now,))
        # if cur.rowcount > 0:
        #     self._logger.debug(f"expired {cur.rowcount} rows")
        # self._conn.commit()

        self._conn.execute('PRAGMA synchronous = OFF')
        self._conn.commit()
        self._conn.close()
        self._conn = sqlite3.connect(path)

    def get(self, key):
        """Retrieve an item."""
        _logger = logging.getLogger(__name__)

        _logger.debug(f"? {key}")
        now = datetime.now().replace(microsecond=0).isoformat()
        cur = self._conn.cursor()
        data = cur.execute("SELECT json FROM items where key=? and expiry > ?", (key, now)).fetchone()
        if data:
            data = json.loads(data[0])
            _logger.debug(f"hit {key}")
        else:
            _logger.debug(f"miss {key}")
        cur.close()
        return data

    def put(self, key, data):
        """Save an item."""
        logging.getLogger(__name__).debug(f"put {key}")
        expiry = (datetime.now() + timedelta(seconds=self._timeout)).replace(microsecond=0).isoformat()
        cur = self._conn.cursor()
        cur.execute("REPLACE into items values (?, ?, ?);", (key, expiry, json.dumps(data, default=json_serial)))
        self._conn.commit()
        cur.close()


def memoize(func):
    """Cache decorator."""

    @functools.wraps(func)
    def memoized_func(*args, **kwargs):
        """Retrieve from cache, or execute func and save results."""
        # special case the first argument (self)
        _id = ''
        if len(args) > 0:
            _id = str(args[0])
            if 'class' in str(type(args[0])):
                if hasattr(args[0], 'id'):
                    _id = args[0].id
        key = func.__name__ + ":" + _id + "/" + "/".join(args[1:]) + "/" + str(kwargs)
        data = cache.get(key)
        if not data:
            data = func(*args, **kwargs)
            cache.put(key, data)
        return data

    return memoized_func


cache = Cache()



"""Represent fhir entity."""


from anvil.transformers.fhir import CANONICAL


class Attachment:
    """Create fhir entity."""

    class_name = "attachment"
    resource_type = "Attachment"

    @staticmethod
    def build_entity(blob):
        """Create fhir entity."""
        # assert False, blob.attributes
        attributes = blob.attributes
        # AttrDict({'size': 32,
        # 'etag': 'CPDS7ffR9+kCEAE=',
        # 'crc32c': 'Lyw+Kw==',
        # 'time_created': '2020-06-10T16:13:14.003000+00:00',
        # 'name': 'gs://fc-secure-004e5c03-d24d-4f7f-a26b-9fdc64b0ca3c/AnVIL_CMG_Broad_Muscle_KNC_WGS_Mar2020/RP-1687/WGS/192CP_ZS_1/v1/192CP_ZS_1.cram.md5',
        # 'property_name': 'md5_path'})
        entity = {
            "resourceType": Attachment.resource_type,
            "id": attributes.etag,
            "extension": [
                {"extension": [
                    {
                        "url": "id",
                        "valueString": "any-id"
                    },
                    {
                        "url": "self_uri",
                        "valueString": "drs://url-here"
                    },
                    {
                        "url": "size",
                        "valueInteger": attributes.size
                    },
                    {
                        "url": "created_time",
                        "valueDateTime": attributes.time_created
                    },
                    {
                        "extension": [
                            {
                                "url": "checksum",
                                "valueString": attributes.crc32c
                            },
                            {
                                "url": "type",
                                "valueString": attributes.etag
                            }
                        ],
                        "url": f"{CANONICAL}/StructureDefinition/drs-checksum"
                    },
                    {
                        "extension": [
                            {
                                "url": "type",
                                "valueString": "gs"
                            },
                            {
                                "url": "access_url",
                                "valueString": attributes.name
                            },
                        ],
                        "url": f"{CANONICAL}/StructureDefinition/drs-access-method"
                    },
                    {
                        "url": "name",
                        "valueString": "any-file-name"
                    },
                    {
                        "url": "mime_type",
                        "valueString": "application/json"
                    }
                ],
                    "url": f"{CANONICAL}/StructureDefinition/drs-object"
                }
            ]
        }
        return entity



"""Represent fhir entity."""

from anvil.transformers.fhir import make_identifier


class Task:
    """Create fhir entity."""

    class_name = "task"
    resource_type = "Task"

    @staticmethod
    def build_entity():
        """Create fhir entity."""
        pass


class SpecimenTask:
    """Create fhir entity linking Specimen to Blobs."""

    class_name = "task"
    resource_type = "Task"

    @staticmethod
    def build_entity(inputs, outputs):
        """Create fhir entity."""
        specimen = inputs[0]
        inputs = [
            {
                "type": {
                    "coding": [
                        {
                            "code": specimen['resourceType']
                        }
                    ]
                },
                "valueReference": {
                    "reference": f"{specimen['resourceType']}/{specimen['id']}"
                }
            }
        ]
        outputs = [
            {
                "type": {
                    "coding": [
                        {
                            "code": blob['resourceType']
                        }
                    ]
                },
                "valueReference": {
                    "reference": f"{blob['resourceType']}/{blob['id']}"
                }
            }
            for blob in outputs]

        return {
            "resourceType": "Task",
            "id": make_identifier("T", specimen['id']),
            "status": "accepted",
            "intent": "unknown",
            "input": inputs,
            "output": outputs,
            "owner": {
                "reference": "Organization/thousandgenomes"
            }
        }
